import contextlib
import copy
import io
import json

import mock
import pytest

from uaclient import event_logger, http, lock, messages, util
from uaclient.cli import main_error_handler
from uaclient.cli.attach import attach_command
from uaclient.exceptions import (
    AlreadyAttachedError,
    LockHeldError,
    MagicAttachInvalidParam,
    MagicAttachTokenError,
    NonRootUserError,
    UbuntuProError,
)
from uaclient.files.user_config_file import UserConfigData
from uaclient.testing.fakes import FakeFile, FakeUbuntuProError
from uaclient.yaml import safe_dump

M_PATH = "uaclient.cli.attach."

# Also used in test_cli_auto_attach.py
BASIC_MACHINE_TOKEN = {
    "availableResources": [],
    "machineToken": "non-empty-token",
    "machineTokenInfo": {
        "machineId": "test_machine_id",
        "contractInfo": {
            "name": "mycontract",
            "id": "contract-1",
            "createdAt": util.parse_rfc3339_date("2020-05-08T19:02:26Z"),
            "effectiveTo": util.parse_rfc3339_date("9999-12-31T00:00:00Z"),
            "resourceEntitlements": [],
            "products": ["free"],
        },
        "accountInfo": {"id": "acct-1", "name": "accountName"},
    },
}

ENTITLED_EXAMPLE_ESM_RESOURCE = {
    "obligations": {"enableByDefault": True},
    "type": "esm-infra",
    "directives": {
        "aptKey": "56F7650A24C9E9ECF87C4D8D4067E40313CB4B13",
        "aptURL": "https://esm.ubuntu.com",
    },
    "affordances": {
        "architectures": [
            "arm64",
            "armhf",
            "i386",
            "ppc64le",
            "s390x",
            "x86_64",
        ],
        "series": ["series-example-1", "series-example-2", "series-example-3"],
    },
    "series": {
        "series-example-1": {
            "directives": {
                "suites": ["example-infra-security", "example-infra-updates"]
            }
        }
    },
    "entitled": True,
}

ENTITLED_MACHINE_TOKEN = copy.deepcopy(BASIC_MACHINE_TOKEN)
ENTITLED_MACHINE_TOKEN["machineTokenInfo"]["contractInfo"][
    "resourceEntitlements"
] = [ENTITLED_EXAMPLE_ESM_RESOURCE]


@mock.patch(
    "uaclient.cli.cli_util.util.we_are_currently_root", return_value=False
)
def test_non_root_users_are_rejected(
    m_we_are_currently_root, FakeConfig, capsys, event
):
    """Check that a UID != 0 will receive a message and exit non-zero"""

    cfg = FakeConfig()
    with pytest.raises(NonRootUserError):
        attach_command.action(mock.MagicMock(), cfg)

    with pytest.raises(SystemExit):
        with mock.patch.object(
            event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
        ):
            main_error_handler(attach_command.action)(mock.MagicMock(), cfg)

    expected = {
        "_schema_version": event_logger.JSON_SCHEMA_VERSION,
        "result": "failure",
        "errors": [
            {
                "message": messages.E_NONROOT_USER.msg,
                "message_code": messages.E_NONROOT_USER.name,
                "service": None,
                "type": "system",
            }
        ],
        "failed_services": [],
        "needs_reboot": False,
        "processed_services": [],
        "warnings": [],
    }
    assert expected == json.loads(capsys.readouterr()[0])


class TestActionAttach:
    def test_already_attached(self, capsys, fake_machine_token_file, event):
        """Check that an already-attached machine emits message and exits 0"""
        fake_machine_token_file.attached = True

        with pytest.raises(AlreadyAttachedError):
            attach_command.action(mock.MagicMock(), cfg=None)

        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                main_error_handler(attach_command.action)(
                    mock.MagicMock(), None
                )

        msg = messages.E_ALREADY_ATTACHED.format(account_name="test")
        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {"account_name": "test"},
                    "message": msg.msg,
                    "message_code": msg.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(capsys.readouterr()[0])

    @mock.patch("uaclient.lock.check_lock_info")
    @mock.patch("time.sleep")
    @mock.patch("uaclient.system.subp")
    def test_lock_file_exists(
        self,
        m_subp,
        m_sleep,
        m_check_lock_info,
        capsys,
        FakeConfig,
        event,
    ):
        cfg = FakeConfig()
        m_check_lock_info.return_value = (123, "pro disable")
        expected_msg = messages.E_LOCK_HELD_ERROR.format(
            lock_request="pro attach", lock_holder="pro disable", pid=123
        )
        """Check when an operation holds a lock file, attach cannot run."""
        with pytest.raises(LockHeldError) as exc_info:
            attach_command.action(mock.MagicMock(), cfg=cfg)
        assert 12 == m_check_lock_info.call_count
        assert expected_msg.msg == exc_info.value.msg

        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                main_error_handler(attach_command.action)(
                    mock.MagicMock(), cfg
                )

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {
                        "lock_holder": "pro disable",
                        "lock_request": "pro attach",
                        "pid": 123,
                    },
                    "message": expected_msg.msg,
                    "message_code": expected_msg.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(capsys.readouterr()[0])

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(
        "uaclient.entitlements.check_entitlement_apt_directives_are_unique",
        return_value=True,
    )
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch("uaclient.files.state_files.machine_id_file.read")
    @mock.patch("uaclient.files.state_files.machine_id_file.write")
    @mock.patch("uaclient.files.state_files.attachment_data_file.write")
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    @mock.patch("uaclient.timer.update_messaging.update_motd_messages")
    @mock.patch(M_PATH + "contract.UAContractClient.add_contract_machine")
    @mock.patch(M_PATH + "cli_util.post_cli_attach")
    def test_happy_path_with_token_arg(
        self,
        m_post_cli,
        contract_machine_attach,
        m_update_apt_and_motd_msgs,
        _m_remove_notice,
        _m_should_reboot,
        _m_attachment_data_file_write,
        _m_machine_id_file_write,
        _m_machine_id_file_read,
        m_update_activity_token,
        _m_check_ent_apt_directives,
        _m_check_lock_info,
        event,
    ):
        """A mock-heavy test for the happy path with the contract token arg"""
        # TODO: Improve this test with less general mocking and more
        # post-conditions
        token = "contract-token"
        args = mock.MagicMock(token=token, attach_config=None)
        cfg = mock.MagicMock()

        def fake_contract_attach(contract_token, attachment_dt):
            return BASIC_MACHINE_TOKEN

        contract_machine_attach.side_effect = fake_contract_attach

        with mock.patch.object(lock, "lock_data_file"):
            ret = attach_command.action(args, cfg)

        assert 0 == ret
        expected_calls = [
            mock.call(contract_token=token, attachment_dt=mock.ANY)
        ]
        assert expected_calls == contract_machine_attach.call_args_list
        assert [mock.call(cfg)] == m_update_apt_and_motd_msgs.call_args_list
        assert 1 == m_update_activity_token.call_count
        assert [mock.call(cfg)] == m_post_cli.call_args_list

    @pytest.mark.parametrize("auto_enable", (True, False))
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch("uaclient.files.state_files.attachment_data_file.write")
    @mock.patch("uaclient.system.should_reboot", return_value=False)
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    @mock.patch("uaclient.status.get_available_resources")
    @mock.patch("uaclient.timer.update_messaging.update_motd_messages")
    @mock.patch(M_PATH + "cli_util.post_cli_attach")
    @mock.patch(M_PATH + "actions.attach_with_token")
    def test_auto_enable_passed_through_to_attach_with_token(
        self,
        m_attach_with_token,
        _m_post_cli_attach,
        m_update_apt_and_motd_msgs,
        _m_get_available_resources,
        _m_remove_notice,
        _m_should_reboot,
        _m_attachment_data_file_write,
        _m_update_activity_token,
        _m_check_lock_info,
        auto_enable,
        FakeConfig,
    ):
        args = mock.MagicMock(
            auto_enable=auto_enable, attach_config=None, token="token"
        )

        cfg = FakeConfig()
        with mock.patch.object(lock, "lock_data_file"):
            attach_command.action(args, cfg)

        assert [
            mock.call(mock.ANY, token="token", allow_enable=auto_enable)
        ] == m_attach_with_token.call_args_list

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    def test_attach_config_and_token_mutually_exclusive(
        self,
        _m_check_lock_info,
        FakeConfig,
    ):
        args = mock.MagicMock(
            token="something", attach_config=FakeFile("something")
        )
        cfg = FakeConfig()
        with pytest.raises(UbuntuProError) as e:
            with mock.patch.object(lock, "lock_data_file"):
                attach_command.action(args, cfg=cfg)

        assert e.value.msg == messages.E_ATTACH_TOKEN_ARG_XOR_CONFIG.msg

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch(M_PATH + "cli_util.post_cli_attach")
    @mock.patch(M_PATH + "actions.attach_with_token")
    def test_token_from_attach_config(
        self,
        m_attach_with_token,
        _m_post_cli_attach,
        m_update_activity_token,
        _m_check_lock_info,
        FakeConfig,
    ):
        args = mock.MagicMock(
            token=None,
            attach_config=FakeFile(safe_dump({"token": "faketoken"})),
        )
        cfg = FakeConfig()
        with mock.patch.object(lock, "lock_data_file"):
            attach_command.action(args, cfg=cfg)

        assert [
            mock.call(mock.ANY, token="faketoken", allow_enable=True)
        ] == m_attach_with_token.call_args_list
        assert 1 == m_update_activity_token.call_count

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    def test_attach_config_invalid_config(
        self,
        _m_check_lock_info,
        FakeConfig,
        capsys,
        event,
    ):
        args = mock.MagicMock(
            token=None,
            attach_config=FakeFile(
                safe_dump({"token": "something", "enable_services": "cis"}),
                name="fakename",
            ),
        )
        cfg = FakeConfig()
        with pytest.raises(UbuntuProError) as e:
            with mock.patch.object(lock, "lock_data_file"):
                attach_command.action(args, cfg=cfg)
        assert "Error while reading fakename:" in e.value.msg

        args.attach_config = FakeFile(
            safe_dump({"token": "something", "enable_services": "cis"}),
            name="fakename",
        )
        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(lock, "lock_data_file"):
                    main_error_handler(attach_command.action)(args, cfg)

        expected_message = messages.E_ATTACH_CONFIG_READ_ERROR.format(
            config_name="fakename",
            error=(
                "Got value with "
                'incorrect type for field "enable_services":\n'
                "Expected value with type list but got type: str"
            ),
        )

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {
                        "config_name": "fakename",
                        "error": (
                            "Got value with "
                            'incorrect type for field "enable_services":\n'
                            "Expected value with type list but got type: str"
                        ),
                    },
                    "message": expected_message.msg,
                    "message_code": expected_message.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(capsys.readouterr()[0])

    @pytest.mark.parametrize("auto_enable", (True, False))
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch(
        M_PATH + "actions.enable_entitlement_by_name",
        return_value=(True, None),
    )
    @mock.patch(M_PATH + "actions.attach_with_token")
    @mock.patch("uaclient.util.handle_unicode_characters")
    @mock.patch("uaclient.status.format_tabular")
    @mock.patch(M_PATH + "actions.status")
    @mock.patch("uaclient.daemon.cleanup")
    @mock.patch("uaclient.daemon.stop")
    def test_attach_config_enable_services(
        self,
        _m_daemon_stop,
        _m_daemon_cleanup,
        m_status,
        m_format_tabular,
        m_handle_unicode,
        m_attach_with_token,
        m_enable,
        m_update_activity_token,
        _m_check_lock_info,
        auto_enable,
        FakeConfig,
        event,
    ):
        m_status.return_value = ("status", 0)
        m_format_tabular.return_value = "status"
        m_handle_unicode.return_value = "status"

        cfg = FakeConfig()
        args = mock.MagicMock(
            token=None,
            attach_config=FakeFile(
                safe_dump({"token": "faketoken", "enable_services": ["cis"]})
            ),
            auto_enable=auto_enable,
        )
        with mock.patch.object(lock, "lock_data_file"):
            attach_command.action(args, cfg=cfg)

        assert [
            mock.call(mock.ANY, token="faketoken", allow_enable=False)
        ] == m_attach_with_token.call_args_list
        if auto_enable:
            assert [mock.call(cfg, "cis")] == m_enable.call_args_list
        else:
            assert [] == m_enable.call_args_list
        assert 1 == m_update_activity_token.call_count

        args.attach_config = FakeFile(
            safe_dump({"token": "faketoken", "enable_services": ["cis"]})
        )

        fake_stdout = io.StringIO()
        with contextlib.redirect_stdout(fake_stdout):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(lock, "lock_data_file"):
                    main_error_handler(attach_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "success",
            "errors": [],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": ["cis"] if auto_enable else [],
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())

    @pytest.mark.parametrize(
        "expected_exception,expected_msg,expected_outer_msg",
        (
            (
                FakeUbuntuProError(),
                messages.E_ATTACH_FAILURE_DEFAULT_SERVICES,
                messages.E_ATTACH_FAILURE_DEFAULT_SERVICES,
            ),
            (
                Exception("error"),
                messages.UNEXPECTED_ERROR.format(
                    error_msg="error",
                    log_path="/var/log/ubuntu-advantage.log",
                ),
                messages.E_ATTACH_FAILURE_UNEXPECTED,
            ),
        ),
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(
        "uaclient.entitlements.check_entitlement_apt_directives_are_unique",
        return_value=True,
    )
    @mock.patch(
        "uaclient.files.state_files.machine_id_file.read", return_value=None
    )
    @mock.patch("uaclient.files.state_files.machine_id_file.write")
    @mock.patch("uaclient.files.state_files.attachment_data_file.write")
    @mock.patch("uaclient.entitlements.entitlements_enable_order")
    @mock.patch("uaclient.actions.enable_entitlement_by_name")
    @mock.patch("uaclient.contract.get_enabled_by_default_services")
    @mock.patch("uaclient.contract.apply_contract_overrides")
    @mock.patch("uaclient.contract.UAContractClient.request_url")
    @mock.patch("uaclient.timer.update_messaging.update_motd_messages")
    @mock.patch(
        "uaclient.files.user_config_file.UserConfigFileObject.public_config",
        new_callable=mock.PropertyMock,
        return_value=UserConfigData(),
    )
    def test_attach_when_one_service_fails_to_enable(
        self,
        m_public_config,
        _m_update_messages,
        m_request_url,
        _m_apply_contract_overrides,
        m_get_enabled_by_default_services,
        m_enable_ent_by_name,
        m_enable_order,
        _m_attachment_data_file_write,
        _m_machine_id_file_write,
        _m_machine_id_file_read,
        _m_check_ent_apt_directives,
        _m_check_lock_info,
        _m_status_cache_file,
        expected_exception,
        expected_msg,
        expected_outer_msg,
        FakeConfig,
        event,
    ):
        args = mock.MagicMock(token="token", attach_config=None)
        cfg = FakeConfig()

        m_enable_order.return_value = ["test1", "test2"]
        m_enable_ent_by_name.side_effect = [
            (True, None),
            expected_exception,
        ]
        m_ent1 = mock.MagicMock(variant="")
        type(m_ent1).name = mock.PropertyMock(return_value="test1")

        m_ent2 = mock.MagicMock(variant="")
        type(m_ent2).name = mock.PropertyMock(return_value="test2")

        m_get_enabled_by_default_services.return_value = [m_ent1, m_ent2]
        m_request_url.return_value = http.HTTPResponse(
            code=200,
            headers={},
            body="",
            json_dict={
                "machineToken": "not-null",
                "machineTokenInfo": {
                    "machineId": "machine-id",
                    "accountInfo": {
                        "id": "acct-1",
                        "name": "acc-name",
                        "createdAt": util.parse_rfc3339_date(
                            "2019-06-14T06:45:50Z"
                        ),
                        "externalAccountIDs": [
                            {"IDs": ["id1"], "origin": "AWS"}
                        ],
                    },
                    "contractInfo": {
                        "id": "cid",
                        "name": "test_contract",
                        "resourceTokens": [
                            {"token": "token", "type": "test1"},
                            {"token": "token", "type": "test2"},
                        ],
                        "resourceEntitlements": [
                            {"type": "test1", "aptURL": "apt"},
                            {"type": "test2", "aptURL": "apt"},
                        ],
                    },
                },
            },
            json_list=[],
        )

        fake_stdout = io.StringIO()
        with pytest.raises(SystemExit):
            with contextlib.redirect_stdout(fake_stdout):
                with mock.patch.object(
                    event,
                    "_event_logger_mode",
                    event_logger.EventLoggerMode.JSON,
                ):
                    with mock.patch.object(lock, "lock_data_file"):
                        main_error_handler(attach_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {
                        "services": [
                            {
                                "code": expected_msg.name,
                                "name": "test2",
                                "title": expected_msg.msg,
                            }
                        ]
                    },
                    "message": expected_outer_msg.msg,
                    "message_code": expected_outer_msg.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": ["test2"],
            "needs_reboot": False,
            "processed_services": ["test1"],
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(M_PATH + "_initiate")
    @mock.patch(M_PATH + "_wait")
    @mock.patch(M_PATH + "_revoke")
    def test_magic_attach_revoke_token_if_wait_fails(
        self,
        m_revoke,
        m_wait,
        m_initiate,
        _m_check_lock_info,
        FakeConfig,
    ):
        m_initiate.return_value = mock.MagicMock(
            token="token", user_code="user_code"
        )
        m_wait.side_effect = MagicAttachTokenError()
        m_args = mock.MagicMock(token=None, attach_config=None)

        with pytest.raises(MagicAttachTokenError):
            with mock.patch.object(lock, "lock_data_file"):
                attach_command.action(args=m_args, cfg=FakeConfig())

        assert 1 == m_initiate.call_count
        assert 1 == m_wait.call_count
        assert 1 == m_revoke.call_count

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    def test_magic_attach_fails_if_format_json_param_used(
        self, _m_check_lock_info, FakeConfig
    ):
        m_args = mock.MagicMock(token=None, attach_config=None, format="json")

        with pytest.raises(MagicAttachInvalidParam) as exc_info:
            with mock.patch.object(lock, "lock_data_file"):
                attach_command.action(args=m_args, cfg=FakeConfig())

        assert (
            "This attach flow does not support --format with value: json"
        ) == exc_info.value.msg



import datetime
from typing import Any, Dict, List, Optional

from uaclient import defaults
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    DatetimeDataValue,
    Field,
    IntDataValue,
    StringDataValue,
    data_list,
)
from uaclient.files.data_types import DataObjectFile, DataObjectFileFormat
from uaclient.files.files import ProJSONFile, UAFile, UserCacheFile
from uaclient.files.user_config_file import LXDGuestAttachEnum

SERVICES_ONCE_ENABLED = "services-once-enabled"


class ServicesOnceEnabledData(DataObject):
    fields = [
        Field("fips_updates", BoolDataValue, False),
    ]

    def __init__(self, fips_updates: bool):
        self.fips_updates = fips_updates


def _services_once_enable_preprocess_data(
    data: Dict[str, Any]
) -> Dict[str, Any]:
    # Since we are using now returning DataObject instances from read, we
    # cannot have variables with "-" in them. We need to explictly convert
    # them before creating the object
    updated_data = {}
    for key in data.keys():
        if "-" in key:
            updated_data[key.replace("-", "_")] = True
        else:
            updated_data[key] = True

    return updated_data


services_once_enabled_file = DataObjectFile(
    data_object_cls=ServicesOnceEnabledData,
    ua_file=UAFile(
        name=SERVICES_ONCE_ENABLED,
        private=False,
    ),
    preprocess_data=_services_once_enable_preprocess_data,
)


class RetryAutoAttachOptions(DataObject):
    fields = [
        Field("enable", data_list(StringDataValue), False),
        Field("enable_beta", data_list(StringDataValue), False),
    ]

    def __init__(
        self,
        enable: Optional[List[str]] = None,
        enable_beta: Optional[List[str]] = None,
    ):
        self.enable = enable
        self.enable_beta = enable_beta


retry_auto_attach_options_file = DataObjectFile(
    RetryAutoAttachOptions,
    UAFile(
        "retry-auto-attach-options.json",
        private=True,
    ),
    DataObjectFileFormat.JSON,
)


class RetryAutoAttachState(DataObject):
    fields = [
        Field("interval_index", IntDataValue),
        Field("failure_reason", StringDataValue, required=False),
    ]

    def __init__(
        self,
        interval_index: int,
        failure_reason: Optional[str],
    ):
        self.interval_index = interval_index
        self.failure_reason = failure_reason


retry_auto_attach_state_file = DataObjectFile(
    RetryAutoAttachState,
    UAFile("retry-auto-attach-state.json", private=True),
    DataObjectFileFormat.JSON,
)


class TimerJobState(DataObject):
    fields = [
        Field("next_run", DatetimeDataValue),
        Field("last_run", DatetimeDataValue),
    ]

    def __init__(self, next_run, last_run):
        self.next_run = next_run
        self.last_run = last_run


class AllTimerJobsState(DataObject):
    fields = [
        Field("metering", TimerJobState, required=False),
        Field("update_messaging", TimerJobState, required=False),
    ]

    def __init__(
        self,
        metering: Optional[TimerJobState],
        update_messaging: Optional[TimerJobState],
    ):
        self.metering = metering
        self.update_messaging = update_messaging


timer_jobs_state_file = DataObjectFile(
    AllTimerJobsState,
    UAFile("jobs-status.json", private=False),
    DataObjectFileFormat.JSON,
)


apt_news_contents_file = UAFile("apt-news", directory=defaults.MESSAGES_DIR)
apt_news_raw_file = UAFile(
    "apt-news-raw", private=False, directory=defaults.MESSAGES_DIR
)


class LivepatchSupportCacheData(DataObject):
    fields = [
        Field("version", StringDataValue),
        Field("flavor", StringDataValue),
        Field("arch", StringDataValue),
        Field("codename", StringDataValue),
        Field("supported", BoolDataValue, required=False),
        Field("cached_at", DatetimeDataValue),
    ]

    def __init__(
        self,
        version: str,
        flavor: str,
        arch: str,
        codename: str,
        supported: Optional[bool],
        cached_at: datetime.datetime,
    ):
        self.version = version
        self.flavor = flavor
        self.arch = arch
        self.codename = codename
        self.supported = supported
        self.cached_at = cached_at


livepatch_support_cache = DataObjectFile(
    LivepatchSupportCacheData,
    UserCacheFile("livepatch-kernel-support-cache.json"),
    file_format=DataObjectFileFormat.JSON,
)

reboot_cmd_marker_file = UAFile("marker-reboot-cmds-required")
only_series_check_marker_file = UAFile("marker-only-series-check")


class AnboxCloudData(DataObject):
    fields = [
        Field("anbox_images_url", StringDataValue),
        Field("anbox_images_resource_token", StringDataValue),
        Field("anbox_cloud_apt_token", StringDataValue),
        Field("anbox_cloud_apt_url", StringDataValue),
    ]

    def __init__(
        self,
        anbox_images_url: str,
        anbox_images_resource_token: str,
        anbox_cloud_apt_token: str,
        anbox_cloud_apt_url: str,
    ):
        self.anbox_images_url = anbox_images_url
        self.anbox_images_resource_token = anbox_images_resource_token
        self.anbox_cloud_apt_token = anbox_cloud_apt_token
        self.anbox_cloud_apt_url = anbox_cloud_apt_url


anbox_cloud_credentials_file = DataObjectFile(
    AnboxCloudData,
    UAFile(
        "anbox-cloud-credentials",
        defaults.DEFAULT_PRIVATE_DATA_DIR,
        private=True,
    ),
    DataObjectFileFormat.JSON,
)


class AttachmentData(DataObject):
    fields = [
        Field("attached_at", DatetimeDataValue),
    ]

    def __init__(self, attached_at: datetime.datetime):
        self.attached_at = attached_at


attachment_data_file = DataObjectFile(
    AttachmentData,
    UAFile("attachment.json", private=False),
    DataObjectFileFormat.JSON,
)


status_cache_file = ProJSONFile(
    pro_file=UAFile(
        name="status.json",
        private=False,
    )
)

machine_id_file = UAFile(
    "machine-id",
    defaults.DEFAULT_PRIVATE_DATA_DIR,
    private=True,
)


class LXDProConfig(DataObject):
    fields = [
        Field("guest_attach", LXDGuestAttachEnum),
    ]

    def __init__(self, guest_attach: LXDGuestAttachEnum):
        self.guest_attach = guest_attach


lxd_pro_config_file = DataObjectFile(
    LXDProConfig,
    UAFile(
        "lxd-config.json",
        directory=defaults.INTERFACE_FILES_DIR,
        private=True,
    ),
    DataObjectFileFormat.JSON,
)


def delete_state_files():
    machine_id_file.delete()
    status_cache_file.delete()
    attachment_data_file.delete()
    anbox_cloud_credentials_file.delete()
    reboot_cmd_marker_file.delete()
    only_series_check_marker_file.delete()
    status_cache_file.delete()
    lxd_pro_config_file.delete()



import logging

from uaclient import exceptions, lock, messages, system, util
from uaclient.api.u.pro.detach.v1 import detach
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def validate_release_series(cfg: UAConfig, only_series, show_message=False):
    LOG.debug("Validating release series")
    if not _is_attached(cfg).is_attached:
        return

    current_series = system.get_release_info().series
    try:
        allowed_release = system.get_distro_info(only_series)
    except exceptions.MissingSeriesInDistroInfoFile:
        # If onlySeries is not present on the distro-info CSV
        # we consider that it is newer than the current release
        pass
    else:
        current_release = system.get_distro_info(current_series)
        # Only series is now meant to be allowed on the specified release
        # and all previous releases
        if current_release.eol > allowed_release.eol:
            LOG.debug(
                "Detaching due to current series %s being higher than only_series: %s",  # noqa
                current_series,
                only_series,
            )
            lock.clear_lock_file_if_present()
            detach()
            message = messages.PRO_ONLY_ALLOWED_FOR_RELEASE.format(
                release=allowed_release.release,
                series_codename=allowed_release.series_codename,
            )
            if show_message:
                print(message)
            LOG.warning(message)



import datetime
import enum
import json
import logging
import re
from functools import lru_cache
from typing import List, Optional, Tuple

from uaclient import event_logger, exceptions, messages, system, util
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    IncorrectTypeError,
    StringDataValue,
    data_list,
)
from uaclient.files import state_files
from uaclient.http import serviceclient

HTTP_PROXY_OPTION = "http-proxy"
HTTPS_PROXY_OPTION = "https-proxy"

LIVEPATCH_CMD = "/snap/bin/canonical-livepatch"

LIVEPATCH_API_V1_KERNELS_SUPPORTED = "/v1/api/kernels/supported"

event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


@enum.unique
class LivepatchSupport(enum.Enum):
    SUPPORTED = object()
    KERNEL_UPGRADE_REQUIRED = object()
    KERNEL_EOL = object()
    UNSUPPORTED = object()
    UNKNOWN = object()


class LivepatchPatchFixStatus(DataObject):
    fields = [
        Field("name", StringDataValue, required=False, dict_key="Name"),
        Field("patched", BoolDataValue, required=False, dict_key="Patched"),
    ]

    def __init__(
        self,
        name: Optional[str],
        patched: Optional[bool],
    ):
        self.name = name
        self.patched = patched


class LivepatchPatchStatus(DataObject):
    fields = [
        Field("state", StringDataValue, required=False, dict_key="State"),
        Field(
            "fixes",
            data_list(LivepatchPatchFixStatus),
            required=False,
            dict_key="Fixes",
        ),
        Field("version", StringDataValue, required=False, dict_key="Version"),
    ]

    def __init__(
        self,
        state: Optional[str],
        fixes: Optional[List[LivepatchPatchFixStatus]],
        version: Optional[str],
    ):
        self.state = state
        self.fixes = fixes
        self.version = version


class LivepatchStatusStatus(DataObject):
    fields = [
        Field("kernel", StringDataValue, required=False, dict_key="Kernel"),
        Field(
            "livepatch",
            LivepatchPatchStatus,
            required=False,
            dict_key="Livepatch",
        ),
        Field(
            "supported",
            StringDataValue,
            required=False,
            dict_key="Supported",
        ),
    ]

    def __init__(
        self,
        kernel: Optional[str],
        livepatch: Optional[LivepatchPatchStatus],
        supported: Optional[str],
    ):
        self.kernel = kernel
        self.livepatch = livepatch
        self.supported = supported


class LivepatchStatus(DataObject):
    fields = [
        Field(
            "status",
            data_list(LivepatchStatusStatus),
            required=False,
            dict_key="Status",
        ),
    ]

    def __init__(
        self,
        status: Optional[List[LivepatchStatusStatus]],
    ):
        self.status = status


def status() -> Optional[LivepatchStatusStatus]:
    if not is_livepatch_installed():
        LOG.debug("canonical-livepatch is not installed")
        return None

    try:
        out, _ = system.subp(
            [LIVEPATCH_CMD, "status", "--verbose", "--format", "json"]
        )
    except exceptions.ProcessExecutionError as e:
        # only raise an error if there is a legitimate problem, not just lack
        # of enablement
        if "Machine is not enabled" in e.stderr:
            LOG.warning(e.stderr)
            return None
        LOG.warning(
            "canonical-livepatch returned error when checking status:\n%s",
            exc_info=e,
        )
        raise e

    try:
        status_json = json.loads(out)
    except json.JSONDecodeError as e:
        LOG.warning(
            "JSONDecodeError while parsing livepatch status, returning None. "
            'output was: "%s"',
            out,
            exc_info=e,
        )
        return None

    try:
        status_root = LivepatchStatus.from_dict(status_json)
    except IncorrectTypeError:
        LOG.warning(
            "canonical-livepatch status returned unexpected structure: %s",
            out,
        )
        return None

    if status_root.status is None or len(status_root.status) < 1:
        LOG.debug("canonical-livepatch has no status")
        return None

    return status_root.status[0]


def _convert_str_to_livepatch_support_status(
    status_str: Optional[str],
) -> Optional[LivepatchSupport]:
    if status_str == "supported":
        return LivepatchSupport.SUPPORTED
    if status_str == "kernel-upgrade-required":
        return LivepatchSupport.KERNEL_UPGRADE_REQUIRED
    if status_str == "kernel-end-of-life":
        return LivepatchSupport.KERNEL_EOL
    if status_str == "unsupported":
        return LivepatchSupport.UNSUPPORTED
    if status_str == "unknown":
        return LivepatchSupport.UNKNOWN
    return None


class UALivepatchClient(serviceclient.UAServiceClient):
    cfg_url_base_attr = "livepatch_url"

    def is_kernel_supported(
        self,
        version: str,
        flavor: str,
        arch: str,
        codename: str,
        build_date: Optional[datetime.datetime],
    ) -> Optional[LivepatchSupport]:
        query_params = {
            "kernel-version": version,
            "flavour": flavor,
            "architecture": arch,
            "codename": codename,
            "build-date": (
                build_date.isoformat() if build_date is not None else "unknown"
            ),
        }
        headers = self.headers()
        try:
            response = self.request_url(
                LIVEPATCH_API_V1_KERNELS_SUPPORTED,
                query_params=query_params,
                headers=headers,
            )
        except Exception as e:
            LOG.warning("error while checking livepatch supported kernels API")
            LOG.warning(e)
            return None

        if response.code != 200:
            LOG.warning("livepatch supported kernels API was unsuccessful")
            LOG.warning(response.body)
            return None

        api_supported_val = response.json_dict.get("Supported")
        if api_supported_val is None or isinstance(api_supported_val, bool):
            # old version, True means supported, None means unsupported
            if api_supported_val:
                return LivepatchSupport.SUPPORTED
            return LivepatchSupport.UNSUPPORTED
        # new version, value is a string
        return _convert_str_to_livepatch_support_status(api_supported_val)


def _on_supported_kernel_cli() -> Optional[LivepatchSupport]:
    try:
        lp_status = status()
    except exceptions.ProcessExecutionError:
        return None

    if lp_status is None:
        return None
    return _convert_str_to_livepatch_support_status(lp_status.supported)


def _on_supported_kernel_cache(
    version: str, flavor: str, arch: str, codename: str
) -> Tuple[bool, Optional[bool]]:
    """Check local cache of kernel support

    :return: (is_cache_valid, result)
    """
    try:
        cache_data = state_files.livepatch_support_cache.read()
    except Exception:
        cache_data = None

    if cache_data is not None:
        one_week_ago = datetime.datetime.now(
            datetime.timezone.utc
        ) - datetime.timedelta(days=7)
        if all(
            [
                cache_data.cached_at > one_week_ago,  # less than one week old
                cache_data.version == version,
                cache_data.flavor == flavor,
                cache_data.arch == arch,
                cache_data.codename == codename,
            ]
        ):
            if cache_data.supported is None:
                LOG.warning("livepatch kernel support cache has None value")
            return (True, cache_data.supported)
    return (False, None)


def _on_supported_kernel_api(
    version: str,
    flavor: str,
    arch: str,
    codename: str,
    build_date: Optional[datetime.datetime],
) -> Optional[LivepatchSupport]:
    supported = UALivepatchClient().is_kernel_supported(
        version=version,
        flavor=flavor,
        arch=arch,
        codename=codename,
        build_date=build_date,
    )

    # cache response as a bool/None before returning
    cache_supported = None
    if supported == LivepatchSupport.SUPPORTED:
        cache_supported = True
    elif supported == LivepatchSupport.UNSUPPORTED:
        cache_supported = False
    state_files.livepatch_support_cache.write(
        state_files.LivepatchSupportCacheData(
            version=version,
            flavor=flavor,
            arch=arch,
            codename=codename,
            supported=cache_supported,
            cached_at=datetime.datetime.now(datetime.timezone.utc),
        )
    )

    if supported is None:
        LOG.warning("livepatch kernel support API response was ambiguous")
    return supported


@lru_cache(maxsize=None)
def on_supported_kernel() -> LivepatchSupport:
    """
    Checks CLI, local cache, and API in that order for kernel support
    If all checks fail to return an authoritative answer, we return None
    """

    # first check cli
    cli_says = _on_supported_kernel_cli()
    if cli_says is not None:
        LOG.debug("using livepatch cli for support")
        return cli_says

    # gather required system info to query support
    kernel_info = system.get_kernel_info()
    if (
        kernel_info.flavor is None
        or kernel_info.major is None
        or kernel_info.minor is None
    ):
        LOG.warning(
            "unable to determine enough kernel information to "
            "check livepatch support"
        )
        return LivepatchSupport.UNKNOWN

    arch = util.standardize_arch_name(kernel_info.uname_machine_arch)
    codename = system.get_release_info().series

    lp_api_kernel_ver = "{major}.{minor}".format(
        major=kernel_info.major, minor=kernel_info.minor
    )

    # second check cache
    is_cache_valid, cache_says = _on_supported_kernel_cache(
        lp_api_kernel_ver, kernel_info.flavor, arch, codename
    )
    if is_cache_valid:
        LOG.debug("using livepatch support cache")
        if cache_says is None:
            return LivepatchSupport.UNKNOWN
        if cache_says:
            return LivepatchSupport.SUPPORTED
        if not cache_says:
            return LivepatchSupport.UNSUPPORTED

    # finally check api
    LOG.debug("using livepatch support api")
    api_says = _on_supported_kernel_api(
        lp_api_kernel_ver,
        kernel_info.flavor,
        arch,
        codename,
        kernel_info.build_date,
    )
    if api_says is None:
        return LivepatchSupport.UNKNOWN
    return api_says


def unconfigure_livepatch_proxy(
    protocol_type: str, retry_sleeps: Optional[List[float]] = None
) -> None:
    """
    Unset livepatch configuration settings for http and https proxies.

    :param protocol_type: String either http or https
    :param retry_sleeps: Optional list of sleep lengths to apply between
        retries. Specifying a list of [0.5, 1] tells subp to retry twice
        on failure; sleeping half a second before the first retry and 1 second
        before the second retry.
    """
    if not is_livepatch_installed():
        return
    system.subp(
        [LIVEPATCH_CMD, "config", "{}-proxy=".format(protocol_type)],
        retry_sleeps=retry_sleeps,
    )


def configure_livepatch_proxy(
    http_proxy: Optional[str] = None,
    https_proxy: Optional[str] = None,
    retry_sleeps: Optional[List[float]] = None,
) -> None:
    """
    Configure livepatch to use http and https proxies.

    :param http_proxy: http proxy to be used by livepatch. If None, it will
                       not be configured
    :param https_proxy: https proxy to be used by livepatch. If None, it will
                        not be configured
    :@param retry_sleeps: Optional list of sleep lengths to apply between
                               snap calls
    """
    from uaclient.entitlements import LivepatchEntitlement

    if http_proxy or https_proxy:
        event.info(
            messages.SETTING_SERVICE_PROXY.format(
                service=LivepatchEntitlement.title
            )
        )

    if http_proxy:
        system.subp(
            [LIVEPATCH_CMD, "config", "http-proxy={}".format(http_proxy)],
            retry_sleeps=retry_sleeps,
        )

    if https_proxy:
        system.subp(
            [LIVEPATCH_CMD, "config", "https-proxy={}".format(https_proxy)],
            retry_sleeps=retry_sleeps,
        )


def get_config_option_value(key: str) -> Optional[str]:
    """
    Gets the config value from livepatch.
    :param key: can be any valid livepatch config option
    :return: the value of the livepatch config option, or None if not set
    """
    out, _ = system.subp([LIVEPATCH_CMD, "config"])
    match = re.search("^{}: (.*)$".format(key), out, re.MULTILINE)
    value = match.group(1) if match else None
    if value:
        # remove quotes if present
        value = re.sub(r"\"(.*)\"", r"\g<1>", value)
    return value.strip() if value else None


def is_livepatch_installed() -> bool:
    return system.which(LIVEPATCH_CMD) is not None



from uaclient.api import exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.contract import UAContractClient
from uaclient.data_types import DataObject, Field, StringDataValue


class MagicAttachRevokeOptions(DataObject):
    fields = [
        Field(
            "magic_token",
            StringDataValue,
            doc="The Token provided by the initiate endpoint.",
        ),
    ]

    def __init__(self, magic_token):
        self.magic_token = magic_token


class MagicAttachRevokeResult(DataObject, AdditionalInfo):
    pass


def revoke(options: MagicAttachRevokeOptions) -> MagicAttachRevokeResult:
    return _revoke(options, UAConfig())


def _revoke(
    options: MagicAttachRevokeOptions, cfg: UAConfig
) -> MagicAttachRevokeResult:
    """
    This endpoint revokes a Magic Attach Token.
    """
    contract = UAContractClient(cfg)
    contract.revoke_magic_attach_token(options.magic_token)

    return MagicAttachRevokeResult()


endpoint = APIEndpoint(
    version="v1",
    name="MagicAttachRevoke",
    fn=_revoke,
    options_cls=MagicAttachRevokeOptions,
)

_doc = {
    "introduced_in": "27.11",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.attach.magic.revoke.v1 import MagicAttachRevokeOptions, revoke

options = MagicAttachWaitOptions(magic_token="<magic_token>")
result = revoke(options)
""",  # noqa: E501
    "result_class": MagicAttachRevokeResult,
    "exceptions": [
        (
            exceptions.ConnectivityError,
            (
                "Raised if it is not possible to connect to the contracts"
                " service."
            ),
        ),
        (
            exceptions.ContractAPIError,
            (
                "Raised if there is an unexpected error in the contracts"
                " service interaction."
            ),
        ),
        (
            exceptions.MagicAttachTokenAlreadyActivated,
            (
                "Raised when trying to revoke a Token which was already"
                " activated through the UI."
            ),
        ),
        (
            exceptions.MagicAttachTokenError,
            "Raised when an invalid/expired Token is sent.",
        ),
        (
            exceptions.MagicAttachUnavailable,
            (
                "Raised if the Magic Attach service is busy or unavailable at"
                " the moment."
            ),
        ),
    ],
    "example_cli": "pro api u.pro.attach.magic.revoke.v1 --args magic_token=<token>",  # noqa: E501
    "example_json": """
{}
""",
}



import mock
import pytest

from uaclient import apt, http, messages
from uaclient.cli import main
from uaclient.cli.cli_util import configure_apt_proxy
from uaclient.cli.config import set_subcommand
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.exceptions import NonRootUserError, UbuntuProError

M_LIVEPATCH = "uaclient.entitlements.livepatch."


@mock.patch("uaclient.log.setup_cli_logging")
class TestMainConfigSet:
    @pytest.mark.parametrize(
        "kv_pair,err_msg",
        (
            ("junk", "\nExpected <key>=<value> but found: junk\n"),
            (
                "k=v",
                "<key> must be one of: http_proxy, https_proxy,"
                " apt_http_proxy, apt_https_proxy, ua_apt_http_proxy,"
                " ua_apt_https_proxy, global_apt_http_proxy,"
                " global_apt_https_proxy, update_messaging_timer,"
                " metering_timer",
            ),
            (
                "http_proxys=",
                "<key> must be one of: http_proxy, https_proxy,"
                " apt_http_proxy, apt_https_proxy, ua_apt_http_proxy,"
                " ua_apt_https_proxy, global_apt_http_proxy,"
                " global_apt_https_proxy, update_messaging_timer,"
                " metering_timer",
            ),
            (
                "=value",
                "<key> must be one of: http_proxy, https_proxy,"
                " apt_http_proxy, apt_https_proxy, ua_apt_http_proxy,"
                " ua_apt_https_proxy, global_apt_http_proxy,"
                " global_apt_https_proxy, update_messaging_timer,"
                " metering_timer",
            ),
            (
                "http_proxy=",
                "Empty value provided for http_proxy.",
            ),
            (
                "https_proxy=  ",
                "Empty value provided for https_proxy.",
            ),
            (
                "apt_news=please",
                "<value> must be one of: true, false",
            ),
        ),
    )
    @mock.patch("uaclient.contract.get_available_resources")
    def test_set_error_with_help_on_invalid_key_value_pair(
        self,
        _m_resources,
        _logging,
        kv_pair,
        err_msg,
        capsys,
        FakeConfig,
        event,
    ):
        """Exit 1 and print help on invalid key_value_pair input param."""
        with pytest.raises(SystemExit):
            with mock.patch(
                "sys.argv", ["/usr/bin/ua", "config", "set", kv_pair]
            ):
                with mock.patch(
                    "uaclient.config.UAConfig",
                    return_value=FakeConfig(),
                ):
                    main()
        out, err = capsys.readouterr()
        assert err_msg in err


@mock.patch("uaclient.config.user_config_file.user_config.write")
@mock.patch("uaclient.contract.get_available_resources")
class TestActionConfigSet:
    @mock.patch("uaclient.util.we_are_currently_root", return_value=False)
    def test_set_error_on_non_root_user(
        self,
        _m_resources,
        _we_are_currently_root,
        _write_,
        FakeConfig,
    ):
        """Root is required to run pro config set."""
        args = mock.MagicMock(key_value_pair="something=1")
        cfg = FakeConfig()
        with pytest.raises(NonRootUserError):
            set_subcommand.action(args, cfg=cfg)

    @pytest.mark.parametrize(
        "key,value,livepatch_enabled",
        (
            ("http_proxy", "http://proxy", False),
            ("https_proxy", "https://proxy", False),
            ("http_proxy", "http://proxy", True),
            ("https_proxy", "https://proxy", True),
        ),
    )
    @mock.patch("uaclient.cli.config.configure_livepatch_proxy")
    @mock.patch(M_LIVEPATCH + "LivepatchEntitlement.application_status")
    @mock.patch("uaclient.cli.config.configure_snap_proxy")
    @mock.patch("uaclient.http.validate_proxy")
    def test_set_http_proxy_and_https_proxy_affects_snap_and_maybe_livepatch(
        self,
        validate_proxy,
        configure_snap_proxy,
        livepatch_status,
        configure_livepatch_proxy,
        _m_resources,
        _write,
        key,
        value,
        livepatch_enabled,
        FakeConfig,
    ):
        """Set updates snap and livepatch proxy, persist config and exits 0.

        Only update livepatch proxy config is livepatch is enabled.
        """
        if livepatch_enabled:
            livepatch_status.return_value = (
                ApplicationStatus.ENABLED,
                "",
            )
        else:
            livepatch_status.return_value = (
                ApplicationStatus.DISABLED,
                "",
            )
        args = mock.MagicMock(key_value_pair="{}={}".format(key, value))
        cfg = FakeConfig()
        set_subcommand.action(args, cfg=cfg)
        kwargs = {key: value}
        if key == "http_proxy":
            url = http.PROXY_VALIDATION_SNAP_HTTP_URL
        else:
            url = http.PROXY_VALIDATION_SNAP_HTTPS_URL
        assert [
            mock.call(key.replace("_proxy", ""), value, url)
        ] == validate_proxy.call_args_list
        assert [mock.call(**kwargs)] == configure_snap_proxy.call_args_list
        if livepatch_enabled:
            assert [
                mock.call(**kwargs)
            ] == configure_livepatch_proxy.call_args_list
        else:
            assert [] == configure_livepatch_proxy.call_args_list

    @pytest.mark.parametrize(
        "key,value,scope,protocol_type",
        (
            (
                "apt_http_proxy",
                "http://proxy",
                apt.AptProxyScope.GLOBAL,
                "http",
            ),
            (
                "apt_https_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                "https",
            ),
        ),
    )
    @mock.patch("uaclient.cli.cli_util.configure_apt_proxy")
    @mock.patch("uaclient.http.validate_proxy")
    def test_set_apt_http_proxy_and_apt_https_proxy_prints_warning(
        self,
        validate_proxy,
        configure_apt_proxy,
        _m_resources,
        _write,
        key,
        value,
        scope,
        protocol_type,
        FakeConfig,
        capsys,
    ):
        """Set calls setup_apt_proxy but prints warning
        and sets global_* and exits 0."""
        args = mock.MagicMock(key_value_pair="{}={}".format(key, value))
        cfg = FakeConfig()
        set_subcommand.action(args, cfg=cfg)
        out, err = capsys.readouterr()
        global_eq = "global_" + key
        assert [
            mock.call(cfg, apt.AptProxyScope.GLOBAL, global_eq, value)
        ] == configure_apt_proxy.call_args_list
        assert (
            messages.WARNING_CONFIG_FIELD_RENAME.format(
                old="apt_{}_proxy".format(protocol_type),
                new="global_apt_{}_proxy".format(protocol_type),
            )
            in out
        )

        proxy_type = key.replace("apt_", "")
        if proxy_type == "http_proxy":
            url = http.PROXY_VALIDATION_APT_HTTP_URL
        else:
            url = http.PROXY_VALIDATION_APT_HTTPS_URL
        assert [
            mock.call(proxy_type.replace("_proxy", ""), value, url)
        ] == validate_proxy.call_args_list

        assert getattr(cfg, global_eq) == value
        assert cfg.ua_apt_https_proxy is None
        assert cfg.ua_apt_http_proxy is None

    @pytest.mark.parametrize(
        "key,value,scope,apt_equ,ua_apt_equ",
        (
            (
                "global_apt_http_proxy",
                "http://proxy",
                apt.AptProxyScope.GLOBAL,
                None,
                None,
            ),
            (
                "global_apt_https_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                "https://proxy",
                None,
            ),
            (
                "global_apt_http_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                "https://proxy",
                None,
            ),
            (
                "global_apt_http_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                None,
                "https://proxy",
            ),
            (
                "global_apt_https_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                None,
                "https://proxy",
            ),
            (
                "global_apt_https_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                "https://proxy",
                "https://proxy",
            ),
            (
                "global_apt_http_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
                "https://proxy",
                "https://proxy",
            ),
        ),
    )
    @mock.patch("uaclient.cli.cli_util.configure_apt_proxy")
    @mock.patch("uaclient.http.validate_proxy")
    def test_set_global_apt_http_and_global_apt_https_proxy(
        self,
        validate_proxy,
        configure_apt_proxy,
        _m_resources,
        _write,
        key,
        value,
        scope,
        apt_equ,
        ua_apt_equ,
        FakeConfig,
        capsys,
    ):
        """Test setting of global_apt_* proxies"""
        args = mock.MagicMock(key_value_pair="{}={}".format(key, value))
        cfg = FakeConfig()
        cfg.ua_apt_https_proxy = ua_apt_equ
        cfg.ua_apt_http_proxy = ua_apt_equ
        set_subcommand.action(args, cfg=cfg)
        out, err = capsys.readouterr()  # will need to check output
        if ua_apt_equ:
            assert [
                mock.call(cfg, scope, key, value)
            ] == configure_apt_proxy.call_args_list
            assert (
                messages.WARNING_APT_PROXY_OVERWRITE.format(
                    current_proxy="global apt", previous_proxy="pro scoped apt"
                )
                in out
            )
        else:
            assert [
                mock.call(cfg, apt.AptProxyScope.GLOBAL, key, value)
            ] == configure_apt_proxy.call_args_list

        proxy_type = key.replace("global_apt_", "")
        if proxy_type == "http_proxy":
            url = http.PROXY_VALIDATION_APT_HTTP_URL
        else:
            url = http.PROXY_VALIDATION_APT_HTTPS_URL
        assert [
            mock.call(proxy_type.replace("_proxy", ""), value, url)
        ] == validate_proxy.call_args_list
        assert cfg.ua_apt_https_proxy is None
        assert cfg.ua_apt_http_proxy is None

    @pytest.mark.parametrize(
        "key,value,scope,apt_equ,global_apt_equ",
        (
            (
                "ua_apt_http_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                None,
                None,
            ),
            (
                "ua_apt_https_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                "https://proxy",
                "https://proxy",
            ),
            (
                "ua_apt_https_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                "https://proxy",
                None,
            ),
            (
                "ua_apt_http_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                "https://proxy",
                None,
            ),
            (
                "ua_apt_https_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                None,
                "https://proxy",
            ),
            (
                "ua_apt_http_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                None,
                "https://proxy",
            ),
            (
                "ua_apt_http_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
                "https://proxy",
                "https://proxy",
            ),
        ),
    )
    @mock.patch("uaclient.cli.cli_util.configure_apt_proxy")
    @mock.patch("uaclient.http.validate_proxy")
    def test_set_ua_apt_http_and_ua_apt_https_proxy(
        self,
        validate_proxy,
        configure_apt_proxy,
        _m_resources,
        _write,
        key,
        value,
        scope,
        apt_equ,
        global_apt_equ,
        FakeConfig,
        capsys,
    ):
        """Test setting of ua_apt_* proxies"""
        args = mock.MagicMock(key_value_pair="{}={}".format(key, value))
        cfg = FakeConfig()
        cfg.global_apt_http_proxy = global_apt_equ
        cfg.global_apt_https_proxy = global_apt_equ
        set_subcommand.action(args, cfg=cfg)
        out, err = capsys.readouterr()  # will need to check output
        if global_apt_equ:
            assert [
                mock.call(cfg, scope, key, value)
            ] == configure_apt_proxy.call_args_list
            assert (
                messages.WARNING_APT_PROXY_OVERWRITE.format(
                    current_proxy="pro scoped apt", previous_proxy="global apt"
                )
                in out
            )
        else:
            assert [
                mock.call(cfg, apt.AptProxyScope.UACLIENT, key, value)
            ] == configure_apt_proxy.call_args_list

        proxy_type = key.replace("ua_apt_", "")
        if proxy_type == "http_proxy":
            url = http.PROXY_VALIDATION_APT_HTTP_URL
        else:
            url = http.PROXY_VALIDATION_APT_HTTPS_URL
        assert [
            mock.call(proxy_type.replace("_proxy", ""), value, url)
        ] == validate_proxy.call_args_list
        assert cfg.global_apt_https_proxy is None
        assert cfg.global_apt_http_proxy is None

    @pytest.mark.parametrize(
        "key,value,scope",
        (
            (
                "global_apt_https_proxy",
                "http://proxy",
                apt.AptProxyScope.GLOBAL,
            ),
            (
                "global_apt_http_proxy",
                "https://proxy",
                apt.AptProxyScope.GLOBAL,
            ),
            ("global_apt_https_proxy", None, apt.AptProxyScope.GLOBAL),
        ),
    )
    @mock.patch("uaclient.cli.cli_util.setup_apt_proxy")
    def test_configure_global_apt_proxy(
        self,
        setup_apt_proxy,
        _m_resources,
        _write,
        key,
        value,
        scope,
        FakeConfig,
    ):
        cfg = FakeConfig()
        cfg.global_apt_http_proxy = value
        cfg.global_apt_https_proxy = value
        configure_apt_proxy(cfg, scope, key, value)
        kwargs = {
            "http_proxy": cfg.global_apt_http_proxy,
            "https_proxy": cfg.global_apt_https_proxy,
            "proxy_scope": scope,
        }
        assert 1 == setup_apt_proxy.call_count
        assert [mock.call(**kwargs)] == setup_apt_proxy.call_args_list

    @pytest.mark.parametrize(
        "key,value,scope",
        (
            (
                "global_apt_https_proxy",
                "http://proxy",
                apt.AptProxyScope.UACLIENT,
            ),
            (
                "global_apt_http_proxy",
                "https://proxy",
                apt.AptProxyScope.UACLIENT,
            ),
            ("global_apt_https_proxy", None, apt.AptProxyScope.UACLIENT),
        ),
    )
    @mock.patch("uaclient.cli.cli_util.setup_apt_proxy")
    def test_configure_uaclient_apt_proxy(
        self,
        setup_apt_proxy,
        _m_resources,
        _write,
        key,
        value,
        scope,
        FakeConfig,
    ):
        cfg = FakeConfig()
        cfg.ua_apt_http_proxy = value
        cfg.ua_apt_https_proxy = value
        configure_apt_proxy(cfg, scope, key, value)
        kwargs = {
            "http_proxy": cfg.ua_apt_http_proxy,
            "https_proxy": cfg.ua_apt_https_proxy,
            "proxy_scope": scope,
        }
        assert 1 == setup_apt_proxy.call_count
        assert [mock.call(**kwargs)] == setup_apt_proxy.call_args_list

    def test_set_timer_interval(self, _m_resources, _write, FakeConfig):
        args = mock.MagicMock(key_value_pair="update_messaging_timer=28800")
        cfg = FakeConfig()
        set_subcommand.action(args, cfg=cfg)
        assert 28800 == cfg.update_messaging_timer

    @pytest.mark.parametrize("invalid_value", ("notanumber", -1))
    def test_error_when_interval_is_not_valid(
        self,
        _m_resources,
        _write,
        FakeConfig,
        invalid_value,
    ):
        args = mock.MagicMock(
            key_value_pair="update_messaging_timer={}".format(invalid_value)
        )
        cfg = FakeConfig()
        with pytest.raises(UbuntuProError):
            set_subcommand.action(args, cfg=cfg)
            assert cfg.update_messaging_timer is None



import json
import logging

import pytest

from lib.patch_status_json import patch_status_json_schema_0_1

UNATTACHED_UNPATCHED_STATUS = {
    "_doc": "Content provided in json response is currently considered ...",
    "attached": False,
    "expires": "n/a",
    "techSupportLevel": "n/a",
    "notices": [],
    "services": [
        {
            "name": "cc-eal",
            "available": "yes",
            "description": "Common Criteria EAL2 Provisioning Packages",
        }
    ],
    "origin": None,
    "configStatusDetails": "No Ubuntu Pro operations are running",
    "configStatus": "inactive",
}

# Note that (UN)ATTACHED_PATCHED_STATUS is only a capture of expected minimal
# schema transformation performed by patch_status function to allow pro status
# to work without errors for non-root user. It is not expected to be a full
# representation of schema which will be written once a root user runs any
# pro command which will rewrite the full latest schema to status.json.
UNATTACHED_PATCHED_STATUS = {
    "execution_details": "No Ubuntu Pro operations are running",
    "_schema_version": "0.1",
    "origin": None,
    "account": {"name": "", "id": ""},
    "expires": "n/a",
    "notices": [],
    "services": [
        {
            "available": "yes",
            "name": "cc-eal",
            "description": "Common Criteria EAL2 Provisioning Packages",
        }
    ],
    "attached": False,
    "execution_status": "inactive",
    "contract": {"name": "", "id": "", "tech_support_level": "n/a"},
    "_doc": "Content provided in json response is currently considered ...",
}

ATTACHED_UNPATCHED_STATUS = {
    "_doc": "Content provided in json response is currently considered ...",
    "configStatusDetails": "Operation in progress: pro attach (pid:33140)",
    "configStatus": "active",
    "subscription": "chad.smith@canonical.com",
    "subscription-id": "cAKuvrqHend2ZHxUYTyJqXhULJ-4r2lpCzd1HLC_lrJg",
    "services": [
        {
            "entitled": "yes",
            "description": "Common Criteria EAL2 Provisioning Packages",
            "description_override": None,
            "name": "cc-eal",
            "statusDetails": "CC EAL2 is not configured",
            "status": "disabled",
        }
    ],
    "notices": [["Operation in progress: pro attach"]],
    "attached": True,
    "account-id": "aAHQlZdfWiafnWvjDZCZDqZzDUComc8WF7IJnoG6GAmA",
    "account": "chad.smith@canonical.com",
    "techSupportLevel": "n/a",
    "expires": "9999-12-31T00:00:00",
    "origin": "free",
}

ATTACHED_PATCHED_STATUS = {
    "_doc": "Content provided in json response is currently considered ...",
    "origin": "free",
    "execution_details": "Operation in progress: pro attach (pid:33140)",
    "execution_status": "active",
    "expires": "9999-12-31T00:00:00+00:00",
    "contract": {
        "tech_support_level": "n/a",
        "name": "chad.smith@canonical.com",
        "id": "cAKuvrqHend2ZHxUYTyJqXhULJ-4r2lpCzd1HLC_lrJg",
    },
    "services": [
        {
            "description_override": None,
            "entitled": "yes",
            "description": "Common Criteria EAL2 Provisioning Packages",
            "status": "disabled",
            "statusDetails": "CC EAL2 is not configured",
            "name": "cc-eal",
        }
    ],
    "_schema_version": "0.1",
    "attached": True,
    "notices": [["Operation in progress: pro attach"]],
    "account": {
        "name": "chad.smith@canonical.com",
        "id": "aAHQlZdfWiafnWvjDZCZDqZzDUComc8WF7IJnoG6GAmA",
    },
}

NO_CONFIG_UNPATCHED_STATUS = {
    "_doc": "Content provided in json response is currently considered ...",
    "attached": False,
    "expires": "n/a",
    "techSupportLevel": "n/a",
    "services": [
        {
            "name": "cc-eal",
            "available": "yes",
            "description": "Common Criteria EAL2 Provisioning Packages",
        }
    ],
    "origin": None,
}

NO_CONFIG_PATCHED_STATUS = {
    "_schema_version": "0.1",
    "origin": None,
    "account": {"name": "", "id": ""},
    "expires": "n/a",
    "services": [
        {
            "available": "yes",
            "name": "cc-eal",
            "description": "Common Criteria EAL2 Provisioning Packages",
        }
    ],
    "attached": False,
    "contract": {"name": "", "id": "", "tech_support_level": "n/a"},
    "_doc": "Content provided in json response is currently considered ...",
}


@pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
class TestPatchStatusJSONSchema0_1:
    @pytest.mark.parametrize(
        "orig_status,expected_status,expected_logs",
        (
            (
                UNATTACHED_UNPATCHED_STATUS,
                UNATTACHED_PATCHED_STATUS,
                ["Patching /var/lib/ubuntu-advantage/status.json schema"],
            ),
            (
                ATTACHED_UNPATCHED_STATUS,
                ATTACHED_PATCHED_STATUS,
                ["Patching /var/lib/ubuntu-advantage/status.json schema"],
            ),
            (
                NO_CONFIG_UNPATCHED_STATUS,
                NO_CONFIG_PATCHED_STATUS,
                ["Patching /var/lib/ubuntu-advantage/status.json schema"],
            ),
            (UNATTACHED_PATCHED_STATUS, UNATTACHED_PATCHED_STATUS, []),
        ),
    )
    def test_unattached_noops(
        self, orig_status, expected_status, expected_logs, tmpdir, caplog_text
    ):
        status_file = tmpdir.join("status.json")
        status_file.write(json.dumps(orig_status))
        patch_status_json_schema_0_1(status_file=status_file.strpath)
        assert expected_status == json.loads(status_file.read())
        debug_logs = caplog_text()
        for log in expected_logs:
            assert log in debug_logs
        if not expected_logs:
            assert (
                "Patching /var/lib/ubuntu-advantage/status.json"
                not in debug_logs
            )



import sys

from uaclient import exceptions, messages
from uaclient.api.u.pro.security.cves.v1 import CVEsOptions, _cves
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.formatter import Table, create_link
from uaclient.cli.parser import HelpCategory
from uaclient.config import UAConfig


@cli_util.with_spinner(msg=messages.CLI_CVES_SPINNER_MSG)
def _get_cve_vulnerabilities(args, *, cfg: UAConfig, **kwargs):
    cve_options = CVEsOptions(
        fixable=args.fixable,
        unfixable=args.unfixable,
    )

    try:
        result = _cves(options=cve_options, cfg=cfg)
    except exceptions.VulnerabilityDataNotFound:
        result = None

    return result


def _get_cve_table_rows(cve_vulnerabilities):
    rows = []

    for package_name, package_info in cve_vulnerabilities.packages.items():
        for cve in package_info.cves:
            cve_info = cve_vulnerabilities.cves.get(cve.name)

            if cve_info:
                rows.append(
                    (
                        package_name,
                        cve_info.priority,
                        cve.fix_origin or "-",
                        cve.name,
                    )
                )

    return rows


def _format_cve_rows(cve_rows):
    formatted_rows = []
    for row in sorted(
        cve_rows,
        key=lambda row: (
            row[0],
            ("critical", "high", "medium", "low", "negligible").index(row[1]),
        ),
    ):
        formatted_rows.append(
            (
                row[0],
                cli_util.colorize_priority(row[1]),
                row[2],
                create_link(
                    text=row[3],
                    url="https://ubuntu.com/security/{}".format(row[3]),
                ),
            )
        )

    return formatted_rows


def _list_cves(args, cfg: UAConfig):
    cve_vulnerabilities = _get_cve_vulnerabilities(args, cfg=cfg)

    if not cve_vulnerabilities:
        raise exceptions.VulnerabilityDataNotFound()

    if cve_vulnerabilities.packages:
        rows = _format_cve_rows(_get_cve_table_rows(cve_vulnerabilities))

        if rows:
            print(
                Table(
                    headers=["Package", "Priority", "Origin", "Vulnerability"],
                    rows=rows,
                )
            )
        else:
            if args.unfixable:
                print(messages.CLI_UNFIXABLE_CVES_NOT_AFFECTED)
            elif args.fixable:
                print(messages.CLI_FIXABLE_CVES_NOT_AFFECTED)
            else:
                print(messages.CLI_CVES_NOT_AFFECTED)
    else:
        if args.unfixable:
            print(messages.CLI_UNFIXABLE_CVES_NOT_AFFECTED)
        elif args.fixable:
            print(messages.CLI_FIXABLE_CVES_NOT_AFFECTED)
        else:
            print(messages.CLI_CVES_NOT_AFFECTED)


def action_cves(args, *, cfg: UAConfig, **kwargs):
    if args.unfixable and args.fixable:
        raise exceptions.InvalidOptionCombination(
            option1="unfixable", option2="fixable"
        )

    try:
        _list_cves(args=args, cfg=cfg)
    except BrokenPipeError:
        sys.stderr.close()


cves_command = ProCommand(
    "cves",
    help=messages.CLI_CVES,
    description=messages.CLI_CVES_DESC,
    action=action_cves,
    help_category=HelpCategory.SECURITY,
    preserve_description=True,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "--unfixable",
                    help=messages.CLI_CVES_UNFIXABLE,
                    action="store_true",
                ),
                ProArgument(
                    "--fixable",
                    help=messages.CLI_CVES_FIXABLE,
                    action="store_true",
                ),
            ]
        )
    ],
)



"""Tests related to uaclient.entitlement.livepatch module."""

import copy
import logging
from functools import partial

import mock
import pytest

from uaclient import apt, exceptions, livepatch, messages
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
    CanEnableFailureReason,
    ContractStatus,
    UserFacingStatus,
)
from uaclient.entitlements.livepatch import (
    LivepatchEntitlement,
    process_config_directives,
)
from uaclient.snap import SNAP_CMD
from uaclient.testing import fakes

M_PATH = "uaclient.entitlements.livepatch."  # mock path
M_LIVEPATCH_STATUS = M_PATH + "LivepatchEntitlement.application_status"
DISABLED_APP_STATUS = (ApplicationStatus.DISABLED, "")

M_BASE_PATH = "uaclient.entitlements.base.UAEntitlement."

DEFAULT_AFFORDANCES = {
    "architectures": ["x86_64"],
    "minKernelVersion": "4.4",
    "kernelFlavors": ["generic", "lowlatency"],
    "tier": "stable",
}


@pytest.fixture
def livepatch_entitlement_factory(entitlement_factory):
    directives = {"caCerts": "", "remoteServer": "https://alt.livepatch.com"}
    return partial(
        entitlement_factory,
        LivepatchEntitlement,
        affordances=DEFAULT_AFFORDANCES,
        directives=directives,
    )


@pytest.fixture
def entitlement(livepatch_entitlement_factory):
    return livepatch_entitlement_factory()


class TestLivepatchContractStatus:
    def test_contract_status_entitled(self, entitlement):
        """The contract_status returns ENTITLED when entitled is True."""
        assert ContractStatus.ENTITLED == entitlement.contract_status()

    def test_contract_status_unentitled(self, livepatch_entitlement_factory):
        """The contract_status returns NONE when entitled is False."""
        entitlement = livepatch_entitlement_factory(entitled=False)
        assert ContractStatus.UNENTITLED == entitlement.contract_status()


class TestLivepatchUserFacingStatus:
    @mock.patch(
        "uaclient.livepatch.on_supported_kernel",
        return_value=None,
    )
    @mock.patch(
        "uaclient.entitlements.livepatch.system.is_container",
        return_value=True,
    )
    def test_user_facing_status_inapplicable_on_inapplicable_status(
        self,
        _m_is_container,
        _m_on_supported_kernel,
        livepatch_entitlement_factory,
    ):
        """The user-facing details INAPPLICABLE applicability_status"""
        affordances = copy.deepcopy(DEFAULT_AFFORDANCES)
        affordances["series"] = ["bionic"]

        entitlement = livepatch_entitlement_factory(affordances=affordances)

        with mock.patch(
            "uaclient.system.get_release_info"
        ) as m_get_release_info:
            m_get_release_info.return_value = mock.MagicMock(series="xenial")
            uf_status, details = entitlement.user_facing_status()
        assert uf_status == UserFacingStatus.INAPPLICABLE
        expected_details = "Cannot install Livepatch on a container."
        assert expected_details == details.msg


class TestLivepatchProcessConfigDirectives:
    @pytest.mark.parametrize(
        "directive_key,livepatch_param_tmpl",
        (("remoteServer", "remote-server={}"), ("caCerts", "ca-certs={}")),
    )
    def test_call_livepatch_config_command(
        self, directive_key, livepatch_param_tmpl
    ):
        """Livepatch config directives are passed to livepatch config."""
        directive_value = "{}-value".format(directive_key)
        cfg = {"entitlement": {"directives": {directive_key: directive_value}}}
        with mock.patch("uaclient.system.subp") as m_subp:
            process_config_directives(cfg)
        expected_subp = mock.call(
            [
                livepatch.LIVEPATCH_CMD,
                "config",
                livepatch_param_tmpl.format(directive_value),
            ],
            capture=True,
        )
        assert [expected_subp] == m_subp.call_args_list

    def test_handle_multiple_directives(self):
        """Handle multiple Livepatch directives using livepatch config."""
        cfg = {
            "entitlement": {
                "directives": {
                    "remoteServer": "value1",
                    "caCerts": "value2",
                    "ignored": "ignoredvalue",
                }
            }
        }
        with mock.patch("uaclient.system.subp") as m_subp:
            process_config_directives(cfg)
        expected_calls = [
            mock.call(
                [livepatch.LIVEPATCH_CMD, "config", "ca-certs=value2"],
                capture=True,
            ),
            mock.call(
                [livepatch.LIVEPATCH_CMD, "config", "remote-server=value1"],
                capture=True,
            ),
        ]
        assert expected_calls == m_subp.call_args_list

    @pytest.mark.parametrize("directives", ({}, {"otherkey": "othervalue"}))
    def test_ignores_other_or_absent(self, directives):
        """Ignore empty or unexpected directives and do not call livepatch."""
        cfg = {"entitlement": {"directives": directives}}
        with mock.patch("uaclient.system.subp") as m_subp:
            process_config_directives(cfg)
        assert 0 == m_subp.call_count


@mock.patch(
    "uaclient.entitlements.fips.FIPSEntitlement.application_status",
    return_value=DISABLED_APP_STATUS,
)
@mock.patch(M_LIVEPATCH_STATUS, return_value=DISABLED_APP_STATUS)
@mock.patch(
    "uaclient.entitlements.livepatch.system.is_container", return_value=False
)
class TestLivepatchEntitlementCanEnable:
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="xenial"),
    )
    @mock.patch(
        "uaclient.system.get_kernel_info",
        return_value=mock.MagicMock(uname_release="4.2.9-00-generic"),
    )
    def test_can_enable_false_on_containers(
        self,
        _m_get_kernel_info,
        _m_get_release_info,
        m_is_container,
        _m_livepatch_status,
        _m_fips_status,
        entitlement,
    ):
        """When is_container is True, can_enable returns False."""
        m_is_container.return_value = True
        entitlement = LivepatchEntitlement(entitlement.cfg)
        result, reason = entitlement.can_enable()
        assert False is result
        assert CanEnableFailureReason.INAPPLICABLE == reason.reason
        msg = "Cannot install Livepatch on a container."
        assert msg == reason.message.msg


class TestLivepatchProcessContractDeltas:
    @mock.patch(M_PATH + "LivepatchEntitlement.setup_livepatch_config")
    def test_true_on_parent_process_deltas(
        self, m_setup_livepatch_config, entitlement
    ):
        """When parent's process_contract_deltas returns True do no setup."""
        assert entitlement.process_contract_deltas({}, {}, False)
        assert [] == m_setup_livepatch_config.call_args_list

    @mock.patch(M_PATH + "LivepatchEntitlement.setup_livepatch_config")
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    @mock.patch(M_PATH + "LivepatchEntitlement.applicability_status")
    def test_false_on_inactive_livepatch_service(
        self,
        m_applicability_status,
        m_application_status,
        m_setup_livepatch_config,
        entitlement,
    ):
        """When livepatch is INACTIVE return False and do no setup."""
        m_applicability_status.return_value = (
            ApplicabilityStatus.APPLICABLE,
            "",
        )
        m_application_status.return_value = (
            ApplicationStatus.DISABLED,
            "",
        )
        deltas = {"entitlement": {"directives": {"caCerts": "new"}}}
        assert not entitlement.process_contract_deltas({}, deltas, False)
        assert [] == m_setup_livepatch_config.call_args_list

    @pytest.mark.parametrize(
        "directives,process_directives,process_token",
        (
            ({"caCerts": "new"}, True, False),
            ({"remoteServer": "new"}, True, False),
            ({"unhandledKey": "new"}, False, False),
        ),
    )
    @mock.patch(M_PATH + "LivepatchEntitlement.setup_livepatch_config")
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    def test_setup_performed_when_active_and_supported_deltas(
        self,
        m_application_status,
        m_setup_livepatch_config,
        entitlement,
        directives,
        process_directives,
        process_token,
    ):
        """Run setup when livepatch ACTIVE and deltas are supported keys."""
        application_status = ApplicationStatus.ENABLED
        m_application_status.return_value = (application_status, "")
        deltas = {"entitlement": {"directives": directives}}
        assert entitlement.process_contract_deltas({}, deltas, False)
        if any([process_directives, process_token]):
            setup_calls = [
                mock.call(
                    progress=mock.ANY,
                    process_directives=process_directives,
                    process_token=process_token,
                )
            ]
        else:
            setup_calls = []
        assert setup_calls == m_setup_livepatch_config.call_args_list

    @pytest.mark.parametrize(
        "deltas,process_directives,process_token",
        (
            ({"entitlement": {"something": 1}}, False, False),
            ({"resourceToken": "new"}, False, True),
        ),
    )
    @mock.patch(M_PATH + "LivepatchEntitlement.setup_livepatch_config")
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    def test_livepatch_disable_and_setup_performed_when_resource_token_changes(
        self,
        m_application_status,
        m_setup_livepatch_config,
        entitlement,
        deltas,
        process_directives,
        process_token,
    ):
        """Run livepatch calls setup when resourceToken changes."""
        application_status = ApplicationStatus.ENABLED
        m_application_status.return_value = (application_status, "")
        entitlement.process_contract_deltas({}, deltas, False)
        if any([process_directives, process_token]):
            setup_calls = [
                mock.call(
                    progress=mock.ANY,
                    process_directives=process_directives,
                    process_token=process_token,
                )
            ]
        else:
            setup_calls = []
        assert setup_calls == m_setup_livepatch_config.call_args_list

    @mock.patch(M_PATH + "LivepatchEntitlement.enable")
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    def test_livepatch_disable_and_delta_with_enable_by_default(
        self,
        m_application_status,
        m_enable,
        entitlement,
    ):
        deltas = {"entitlement": {"obligations": {"enableByDefault": True}}}
        m_enable.return_value = (True, None)
        application_status = ApplicationStatus.DISABLED
        m_application_status.return_value = (application_status, "")
        entitlement.process_contract_deltas({}, deltas, False)
        assert [mock.call(mock.ANY)] == m_enable.call_args_list


@mock.patch(M_PATH + "snap.is_snapd_installed_as_a_snap")
@mock.patch(M_PATH + "snap.is_snapd_installed")
@mock.patch("uaclient.http.validate_proxy", side_effect=lambda x, y, z: y)
@mock.patch("uaclient.snap.configure_snap_proxy")
@mock.patch("uaclient.livepatch.configure_livepatch_proxy")
class TestLivepatchEntitlementEnable:

    mocks_apt_update = []
    mocks_snapd_install = [
        mock.call(
            ["apt-get", "install", "--assume-yes", "snapd"],
            retry_sleeps=apt.APT_RETRIES,
        )
    ]
    mocks_snapd_install_as_a_snap = [
        mock.call(
            ["/usr/bin/snap", "install", "snapd"],
            capture=True,
            retry_sleeps=[0.5, 1, 5],
        )
    ]
    mocks_snap_wait_seed = [
        mock.call(
            ["/usr/bin/snap", "wait", "system", "seed.loaded"], capture=True
        )
    ]
    mocks_snapd_refresh = [
        mock.call(
            ["/usr/bin/snap", "refresh", "snapd"],
            capture=True,
        )
    ]
    mocks_livepatch_install = [
        mock.call(
            ["/usr/bin/snap", "install", "canonical-livepatch"],
            capture=True,
            retry_sleeps=[0.5, 1, 5],
        )
    ]
    mocks_install = (
        mocks_snapd_install
        + mocks_snapd_install_as_a_snap
        + mocks_snap_wait_seed
        + mocks_snapd_refresh
        + mocks_livepatch_install
    )
    mocks_config = [
        mock.call(
            [
                livepatch.LIVEPATCH_CMD,
                "config",
                "remote-server=https://alt.livepatch.com",
            ],
            capture=True,
        ),
        mock.call([livepatch.LIVEPATCH_CMD, "disable"]),
        mock.call(
            [livepatch.LIVEPATCH_CMD, "enable", "livepatch-token"],
            capture=True,
        ),
    ]

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize("apt_update_success", (True, False))
    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.contract.apply_contract_overrides")
    @mock.patch("uaclient.apt.update_sources_list")
    @mock.patch("uaclient.apt.run_apt_install_command")
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.system.which", return_value=None)
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    @mock.patch(
        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
    )
    def test_enable_installs_snapd_and_livepatch_snap_when_absent(
        self,
        m_can_enable,
        m_app_status,
        m_which,
        m_run_apt_update,
        m_run_apt_install,
        m_update_sources_list,
        _m_contract_overrides,
        m_subp,
        _m_get_release_info,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        capsys,
        caplog_text,
        event,
        entitlement,
        apt_update_success,
    ):
        """Install snapd and canonical-livepatch snap when not on system."""
        application_status = ApplicationStatus.ENABLED
        m_app_status.return_value = application_status, "enabled"
        m_is_snapd_installed.return_value = False
        m_is_snapd_installed_as_a_snap.return_value = False

        def fake_update_sources_list(sources_list):
            if apt_update_success:
                return
            raise fakes.FakeUbuntuProError()

        m_update_sources_list.side_effect = fake_update_sources_list

        progress_mock = mock.MagicMock()
        assert entitlement.enable(progress_mock)
        assert self.mocks_install + self.mocks_config in m_subp.call_args_list
        assert self.mocks_apt_update == m_run_apt_update.call_args_list
        assert 1 == m_update_sources_list.call_count
        assert [
            mock.call("Installing Livepatch"),
            mock.call("Setting up Livepatch"),
        ] == progress_mock.progress.call_args_list
        assert [
            mock.call("message_operation", None),
            mock.call("message_operation", None),
            mock.call("info", "Installing snapd"),
            mock.call("info", "Installing snapd snap"),
            mock.call("info", "Installing canonical-livepatch snap"),
            mock.call(
                "info", "Disabling Livepatch prior to re-attach with new token"
            ),
        ] == progress_mock.emit.call_args_list
        assert [mock.call(livepatch.LIVEPATCH_CMD)] == m_which.call_args_list
        expected_log = (
            "DEBUG    Trying to install snapd."
            " Ignoring apt-get update failure: This is a test"
        )
        if apt_update_success:
            assert expected_log not in caplog_text()
        else:
            assert expected_log in caplog_text()
        assert [mock.call(livepatch.LIVEPATCH_CMD)] == m_which.call_args_list
        assert m_validate_proxy.call_count == 2
        assert m_snap_proxy.call_count == 1
        assert m_livepatch_proxy.call_count == 1

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.contract.apply_contract_overrides")
    @mock.patch("uaclient.apt.update_sources_list")
    @mock.patch("uaclient.apt.run_apt_install_command")
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.system.which", return_value=None)
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    @mock.patch(
        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
    )
    def test_enable_continues_when_snap_install_snapd_fails(
        self,
        m_can_enable,
        m_app_status,
        m_which,
        m_run_apt_update,
        m_run_apt_install,
        m_update_sources_list,
        _m_contract_overrides,
        m_subp,
        _m_get_release_info,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        capsys,
        caplog_text,
        event,
        entitlement,
    ):
        """Install snapd and canonical-livepatch snap when not on system."""
        application_status = ApplicationStatus.ENABLED
        m_app_status.return_value = application_status, "enabled"
        m_is_snapd_installed.return_value = False
        m_is_snapd_installed_as_a_snap.return_value = False
        m_subp.side_effect = [
            None,
            exceptions.ProcessExecutionError("test"),
            None,
            None,
            None,
            None,
            None,
            None,
        ]

        progress_mock = mock.MagicMock()
        assert entitlement.enable(progress_mock)
        assert self.mocks_install + self.mocks_config in m_subp.call_args_list
        assert self.mocks_apt_update == m_run_apt_update.call_args_list
        assert 1 == m_update_sources_list.call_count
        assert [
            mock.call("Installing Livepatch"),
            mock.call("Setting up Livepatch"),
        ] == progress_mock.progress.call_args_list
        assert [
            mock.call("message_operation", None),
            mock.call("message_operation", None),
            mock.call("info", "Installing snapd"),
            mock.call("info", "Installing snapd snap"),
            mock.call("info", "Executing `snap install snapd` failed."),
            mock.call("info", "Installing canonical-livepatch snap"),
            mock.call(
                "info", "Disabling Livepatch prior to re-attach with new token"
            ),
        ] == progress_mock.emit.call_args_list
        assert [mock.call(livepatch.LIVEPATCH_CMD)] == m_which.call_args_list
        assert m_validate_proxy.call_count == 2
        assert m_snap_proxy.call_count == 1
        assert m_livepatch_proxy.call_count == 1

    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.contract.apply_contract_overrides")
    @mock.patch("uaclient.system.which", return_value=None)
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    @mock.patch(
        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
    )
    def test_enable_installs_only_livepatch_snap_when_absent_but_snapd_present(
        self,
        _m_can_enable,
        m_app_status,
        m_which,
        _m_contract_overrides,
        m_subp,
        _m_get_release_info,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        capsys,
        event,
        entitlement,
    ):
        """Install canonical-livepatch snap when not present on the system."""
        application_status = ApplicationStatus.ENABLED
        m_app_status.return_value = application_status, "enabled"
        m_is_snapd_installed.return_value = True

        progress_mock = mock.MagicMock()
        assert entitlement.enable(progress_mock)
        assert (
            self.mocks_snap_wait_seed
            + self.mocks_snapd_refresh
            + self.mocks_livepatch_install
            + self.mocks_config
            in m_subp.call_args_list
        )
        assert [
            mock.call("Installing Livepatch"),
            mock.call("Setting up Livepatch"),
        ] == progress_mock.progress.call_args_list
        assert [
            mock.call("message_operation", None),
            mock.call("message_operation", None),
            mock.call("info", "Installing canonical-livepatch snap"),
            mock.call(
                "info", "Disabling Livepatch prior to re-attach with new token"
            ),
        ] == progress_mock.emit.call_args_list
        assert [mock.call(livepatch.LIVEPATCH_CMD)] == m_which.call_args_list
        assert m_validate_proxy.call_count == 2
        assert m_snap_proxy.call_count == 1
        assert m_livepatch_proxy.call_count == 1

    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.contract.apply_contract_overrides")
    @mock.patch(
        "uaclient.system.which", side_effect=["/path/to/exe", "/path/to/exe"]
    )
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    @mock.patch(
        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
    )
    def test_enable_does_not_install_livepatch_snap_when_present(
        self,
        m_can_enable,
        m_app_status,
        m_which,
        _m_contract_overrides,
        m_subp,
        _m_get_release_info,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        capsys,
        event,
        entitlement,
    ):
        """Do not attempt to install livepatch snap when it is present."""
        application_status = ApplicationStatus.ENABLED
        m_app_status.return_value = application_status, "enabled"
        m_is_snapd_installed.return_value = True

        progress_mock = mock.MagicMock()
        assert entitlement.enable(progress_mock)
        subp_calls = [
            mock.call(
                [SNAP_CMD, "wait", "system", "seed.loaded"], capture=True
            ),
            mock.call([SNAP_CMD, "refresh", "snapd"], capture=True),
            mock.call(
                [
                    livepatch.LIVEPATCH_CMD,
                    "config",
                    "remote-server=https://alt.livepatch.com",
                ],
                capture=True,
            ),
            mock.call([livepatch.LIVEPATCH_CMD, "disable"]),
            mock.call(
                [livepatch.LIVEPATCH_CMD, "enable", "livepatch-token"],
                capture=True,
            ),
        ]
        assert subp_calls == m_subp.call_args_list
        assert [
            mock.call("message_operation", None),
            mock.call("message_operation", None),
            mock.call(
                "info", "Disabling Livepatch prior to re-attach with new token"
            ),
        ] == progress_mock.emit.call_args_list
        assert m_validate_proxy.call_count == 2
        assert m_snap_proxy.call_count == 1
        assert m_livepatch_proxy.call_count == 1

    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.contract.apply_contract_overrides")
    @mock.patch(
        "uaclient.system.which", side_effect=["/path/to/exe", "/path/to/exe"]
    )
    @mock.patch(M_PATH + "LivepatchEntitlement.application_status")
    @mock.patch(
        M_PATH + "LivepatchEntitlement.can_enable", return_value=(True, None)
    )
    def test_enable_does_not_disable_inactive_livepatch_snap_when_present(
        self,
        m_can_enable,
        m_app_status,
        m_which,
        _m_contract_overrides,
        m_subp,
        _m_get_release_info,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        capsys,
        entitlement,
    ):
        """Do not attempt to disable livepatch snap when it is inactive."""
        m_app_status.return_value = ApplicationStatus.DISABLED, "nope"
        m_is_snapd_installed.return_value = True

        assert entitlement.enable(mock.MagicMock())
        subp_no_livepatch_disable = [
            mock.call(
                [SNAP_CMD, "wait", "system", "seed.loaded"], capture=True
            ),
            mock.call([SNAP_CMD, "refresh", "snapd"], capture=True),
            mock.call(
                [
                    livepatch.LIVEPATCH_CMD,
                    "config",
                    "remote-server=https://alt.livepatch.com",
                ],
                capture=True,
            ),
            mock.call(
                [livepatch.LIVEPATCH_CMD, "enable", "livepatch-token"],
                capture=True,
            ),
        ]
        assert subp_no_livepatch_disable == m_subp.call_args_list
        assert m_validate_proxy.call_count == 2
        assert m_snap_proxy.call_count == 1
        assert m_livepatch_proxy.call_count == 1

    @pytest.mark.parametrize("caplog_text", [logging.WARN], indirect=True)
    @mock.patch("uaclient.system.which", return_value=None)
    @mock.patch("uaclient.system.subp")
    def test_enable_alerts_user_that_snapd_does_not_wait_command(
        self,
        m_subp,
        _m_which,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        entitlement,
        capsys,
        caplog_text,
        event,
    ):
        m_is_snapd_installed.return_value = True

        stderr_msg = (
            "error: Unknown command `wait'. Please specify one command of: "
            "abort, ack, buy, change, changes, connect, create-user, disable,"
            " disconnect, download, enable, find, help, install, interfaces, "
            "known, list, login, logout, refresh, remove, run or try"
        )

        m_subp.side_effect = [
            exceptions.ProcessExecutionError(
                cmd="snapd wait system seed.loaded",
                exit_code=-1,
                stdout="",
                stderr=stderr_msg,
            ),
            True,
            True,
        ]

        progress_mock = mock.MagicMock()

        with mock.patch.object(entitlement, "can_enable") as m_can_enable:
            m_can_enable.return_value = (True, None)
            with mock.patch.object(
                entitlement, "setup_livepatch_config"
            ) as m_setup_livepatch:
                entitlement.enable(progress_mock)

                assert 1 == m_can_enable.call_count
                assert 1 == m_setup_livepatch.call_count

        assert (
            mock.call("info", "Installing canonical-livepatch snap")
            in progress_mock.emit.call_args_list
        )

        assert (
            "Detected version of snapd that does not have wait command"
            in caplog_text()
        )

        assert m_validate_proxy.call_count == 2
        assert m_snap_proxy.call_count == 1
        assert m_livepatch_proxy.call_count == 1

    @mock.patch("uaclient.snap.apt.run_apt_update_command")
    @mock.patch("uaclient.system.which", return_value=True)
    @mock.patch("uaclient.system.subp")
    def test_enable_raise_exception_when_snapd_cant_be_installed(
        self,
        m_subp,
        _m_which,
        _m_apt_update,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        entitlement,
    ):
        m_is_snapd_installed.return_value = False
        m_subp.side_effect = exceptions.ProcessExecutionError(
            cmd="apt-get install --assume-yes snapd",
            exit_code=-1,
        )

        with mock.patch.object(entitlement, "can_enable") as m_can_enable:
            m_can_enable.return_value = (True, None)
            with mock.patch.object(
                entitlement, "setup_livepatch_config"
            ) as m_setup_livepatch:
                with pytest.raises(exceptions.CannotInstallSnapdError):
                    entitlement.enable(mock.MagicMock())

            assert 1 == m_can_enable.call_count
            assert 0 == m_setup_livepatch.call_count

        assert m_validate_proxy.call_count == 0
        assert m_snap_proxy.call_count == 0
        assert m_livepatch_proxy.call_count == 0

    @mock.patch("uaclient.snap.apt.run_apt_update_command")
    @mock.patch("uaclient.system.which", return_value="/path/to/exe")
    @mock.patch("uaclient.system.subp")
    def test_enable_raise_exception_for_unexpected_error_on_snapd_wait(
        self,
        m_subp,
        _m_which,
        _m_apt_update,
        m_livepatch_proxy,
        m_snap_proxy,
        m_validate_proxy,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        entitlement,
    ):
        m_is_snapd_installed.return_value = True
        stderr_msg = "test error"

        m_subp.side_effect = exceptions.ProcessExecutionError(
            cmd="snapd wait system seed.loaded",
            exit_code=-1,
            stdout="",
            stderr=stderr_msg,
        )

        with mock.patch.object(entitlement, "can_enable") as m_can_enable:
            m_can_enable.return_value = (True, None)
            with mock.patch.object(
                entitlement, "setup_livepatch_config"
            ) as m_setup_livepatch:
                with pytest.raises(
                    exceptions.ProcessExecutionError
                ) as excinfo:
                    entitlement.enable(mock.MagicMock())

            assert 1 == m_can_enable.call_count
            assert 0 == m_setup_livepatch.call_count

        expected_msg = "test error"
        assert expected_msg in str(excinfo)
        assert m_validate_proxy.call_count == 0
        assert m_snap_proxy.call_count == 0
        assert m_livepatch_proxy.call_count == 0


class TestLivepatchApplicationStatus:
    @pytest.mark.parametrize("which_result", (("/path/to/exe"), (None)))
    @pytest.mark.parametrize(
        "livepatch_status_result",
        (
            (None),
            (
                livepatch.LivepatchStatusStatus(
                    kernel=None, livepatch=None, supported=None
                )
            ),
        ),
    )
    @mock.patch("uaclient.system.which")
    @mock.patch("uaclient.livepatch.status")
    def test_application_status(
        self,
        m_livepatch_status,
        m_which,
        livepatch_status_result,
        which_result,
        entitlement,
    ):
        m_which.return_value = which_result
        m_livepatch_status.return_value = livepatch_status_result

        status, details = entitlement.application_status()

        if not which_result:
            assert status == ApplicationStatus.DISABLED
            assert "canonical-livepatch snap is not installed." in details.msg
        elif livepatch_status_result is None:
            assert status == ApplicationStatus.DISABLED
            assert (
                messages.LIVEPATCH_APPLICATION_STATUS_CLIENT_FAILURE == details
            )
        else:
            assert status == ApplicationStatus.ENABLED
            assert details is None

    @mock.patch("uaclient.livepatch.is_livepatch_installed", return_value=True)
    @mock.patch("uaclient.livepatch.status")
    def test_application_status_when_canonical_livepatch_fails(
        self, m_status, _m_livepatch_installed, entitlement
    ):
        m_status.side_effect = exceptions.ProcessExecutionError(
            cmd="test", stdout="", stderr="livepatch error"
        )

        status, details = entitlement.application_status()

        assert status == ApplicationStatus.WARNING
        assert details == messages.LIVEPATCH_CLIENT_FAILURE_WARNING.format(
            livepatch_error="livepatch error"
        )



import json

from uaclient import config, entitlements, messages, status
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand


def action_help(args, *, cfg, **kwargs):
    service = args.service

    if not service:
        # Avoiding a circular import
        from uaclient.cli import get_parser

        get_parser().print_help()
        return 0

    if not cfg:
        cfg = config.UAConfig()

    help_response = status.help(cfg, service)

    if args.format == "json":
        print(json.dumps(help_response))
    else:
        for key, value in help_response.items():
            print("{}:\n{}\n".format(key.title(), value))

    return 0


help_command = ProCommand(
    "help",
    help=messages.CLI_ROOT_HELP,
    description=messages.CLI_HELP_DESC,
    action=action_help,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "service",
                    help=messages.CLI_HELP_SERVICE.format(
                        options=", ".join(
                            entitlements.valid_services(cfg=config.UAConfig())
                        )
                    ),
                    action="store",
                    nargs="?",
                ),
                ProArgument(
                    "--format",
                    help=(messages.CLI_FORMAT_DESC.format(default="tabular")),
                    action="store",
                    choices=["tabular", "json", "yaml"],
                    default="tabular",
                ),
                ProArgument(
                    "--all", help=messages.CLI_HELP_ALL, action="store_true"
                ),
            ]
        )
    ],
)



import logging

from uaclient import config, defaults, log
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.entitlements.esm import ESMAppsEntitlement, ESMInfraEntitlement

LOG = logging.getLogger("ubuntupro.lib.add_esm_snapshot_auth")

ESM_APPS_SNAPSHOT_URLS = [
    "snapshot.apps-security.esm.ubuntu.com/apps/ubuntu/",
    "snapshot.apps-updates.esm.ubuntu.com/apps/ubuntu/",
]
ESM_INFRA_SNAPSHOT_URLS = [
    "snapshot.infra-security.esm.ubuntu.com/infra/ubuntu/",
    "snapshot.infra-updates.esm.ubuntu.com/infra/ubuntu/",
]


def add_esm_snapshot_auth(cfg):
    if not _is_attached(cfg).is_attached:
        LOG.info("Not attached. Not adding ESM snapshot URLs to APT auth.")
        return
    for esm, urls in [
        (ESMAppsEntitlement(cfg), ESM_APPS_SNAPSHOT_URLS),
        (ESMInfraEntitlement(cfg), ESM_INFRA_SNAPSHOT_URLS),
    ]:
        if (
            esm.application_status()[0] != ApplicationStatus.DISABLED
            and esm.apt_url is not None
            and "https://esm.ubuntu.com/" in esm.apt_url
        ):
            LOG.info(
                (
                    "%s is enabled and using default apt url,"
                    " adding snapshot auth"
                ),
                esm.name,
            )
            esm.update_apt_auth(override_snapshot_urls=urls)
        else:
            LOG.info(
                "%s is disabled or not using default apt url, skipping",
                esm.name,
                extra=log.extra(apt_url=esm.apt_url),
            )


if __name__ == "__main__":
    try:
        log.setup_cli_logging(
            logging.DEBUG, defaults.CONFIG_DEFAULTS["log_level"]
        )
        cfg = config.UAConfig()
        log.setup_cli_logging(cfg.log_level, cfg.log_file)
        add_esm_snapshot_auth(cfg)
    except Exception as e:
        print(e)



import os.path

import mock
import pytest

from uaclient.version import (
    check_for_new_version,
    get_last_known_candidate,
    get_version,
)


@mock.patch("uaclient.version.subp")
class TestGetVersion:
    @mock.patch("uaclient.version.os.path.exists", return_value=True)
    def test_get_version_returns_packaged_version(self, m_exists, m_subp):
        with mock.patch("uaclient.version.PACKAGED_VERSION", "24.1~18.04.1"):
            assert "24.1~18.04.1" == get_version()
        assert 0 == m_subp.call_count

    @mock.patch("uaclient.version.os.path.exists", return_value=True)
    def test_get_version_returns_matching_git_describe_long(
        self, m_exists, m_subp
    ):
        m_subp.return_value = ("24.1-5-g12345678", "")
        with mock.patch(
            "uaclient.version.PACKAGED_VERSION", "@@PACKAGED_VERSION"
        ):
            assert "24.1-5-g12345678" == get_version()
        assert [
            mock.call(
                ["git", "describe", "--abbrev=8", "--match=[0-9]*", "--long"]
            )
        ] == m_subp.call_args_list
        top_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        top_dir_git = os.path.join(top_dir, ".git")
        assert [mock.call(top_dir_git)] == m_exists.call_args_list


class TestGetLastKnownCandidate:
    @mock.patch("builtins.open", mock.mock_open(read_data="1.2.3"))
    @mock.patch("os.stat")
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("uaclient.version.get_apt_cache_time", return_value=1)
    def test_get_known_candidate_from_cache(
        self, _m_apt_cache_time, _m_exists, m_stat
    ):
        m_stat.return_value.st_mtime = 2
        assert "1.2.3" == get_last_known_candidate()

    @mock.patch("builtins.open")
    @mock.patch("os.stat")
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("uaclient.version.get_apt_cache_time", return_value=1)
    def test_cant_open_cache_file(
        self, _m_apt_cache_time, _m_exists, m_stat, m_open
    ):
        m_stat.return_value.st_mtime = 2
        m_open.side_effect = OSError()
        assert None is get_last_known_candidate()

    @mock.patch("os.makedirs")
    @mock.patch(
        "uaclient.version.get_pkg_candidate_version", return_value="1.2.3"
    )
    @mock.patch("os.path.exists", return_value=False)
    def test_create_cache_before_returning(
        self, _m_exists, _m_candidate, _m_makedirs
    ):
        with mock.patch("builtins.open", mock.mock_open()) as m_open:
            get_last_known_candidate()
            assert m_open.return_value.write.call_args_list == [
                mock.call("1.2.3")
            ]

    @mock.patch("builtins.open")
    @mock.patch(
        "uaclient.version.get_pkg_candidate_version", return_value="1.2.3"
    )
    @mock.patch("os.path.exists", return_value=False)
    def test_problem_updating_file(self, _m_exists, _m_candidate, m_open):
        m_open.side_effect = OSError()
        assert "1.2.3" == get_last_known_candidate()


class TestCheckForNewVersion:
    @pytest.mark.parametrize("compare_return", (-1, 0, 1))
    @mock.patch("uaclient.version.version_compare")
    @mock.patch("uaclient.version.get_last_known_candidate")
    def test_check_for_new_version(
        self, m_candidate, m_compare, compare_return
    ):
        m_compare.return_value = compare_return
        if compare_return == 1:
            assert m_candidate.return_value == check_for_new_version()
        else:
            assert None is check_for_new_version()



import mock
import pytest

from uaclient import exceptions, messages
from uaclient.api.data_types import ErrorWarningObject
from uaclient.api.u.pro.attach.token.full_token_attach.v1 import (
    FullTokenAttachResult,
    _full_token_attach,
)

M_PATH = "uaclient.api.u.pro.attach.token.full_token_attach.v1."


@mock.patch("uaclient.lock.RetryLock.__enter__")
class TestFullTokenAttach:
    @mock.patch("uaclient.util.we_are_currently_root")
    def test_full_token_attach_non_root_user(
        self, m_we_are_currently_root, _m_lock_enter
    ):
        m_we_are_currently_root.return_value = False

        with pytest.raises(exceptions.NonRootUserError):
            _full_token_attach(None, None)

    @mock.patch(M_PATH + "_is_attached")
    def test_full_token_attach_when_already_attached(
        self, m_is_attached, _m_lock_enter
    ):
        m_is_attached.return_value = mock.MagicMock(is_attached=True)
        assert FullTokenAttachResult(
            enabled=[], reboot_required=False
        ) == _full_token_attach(None, None)

    @pytest.mark.parametrize(
        "test_exception,expected_result,expected_warnings",
        (
            (
                exceptions.AttachFailureDefaultServices(
                    failed_services=[
                        ("ent1", messages.E_ATTACH_FAILURE_DEFAULT_SERVICES),
                        ("ent2", messages.E_ATTACH_FAILURE_DEFAULT_SERVICES),
                    ]
                ),
                FullTokenAttachResult(
                    enabled=["ent3"],
                    reboot_required=True,
                ),
                [
                    ErrorWarningObject(
                        title=messages.E_ATTACH_FAILURE_DEFAULT_SERVICES.msg,
                        code=messages.E_ATTACH_FAILURE_DEFAULT_SERVICES.name,
                        meta={"service": "ent1"},
                    ),
                    ErrorWarningObject(
                        title=messages.E_ATTACH_FAILURE_DEFAULT_SERVICES.msg,
                        code=messages.E_ATTACH_FAILURE_DEFAULT_SERVICES.name,
                        meta={"service": "ent2"},
                    ),
                ],
            ),
            (
                exceptions.AttachFailureUnknownError(
                    failed_services=[
                        (
                            "ent1",
                            messages.UNEXPECTED_ERROR.format(
                                error_msg="error",
                                log_path="path",
                            ),
                        )
                    ]
                ),
                FullTokenAttachResult(
                    enabled=["ent3"],
                    reboot_required=True,
                ),
                [
                    ErrorWarningObject(
                        title=messages.UNEXPECTED_ERROR.format(
                            error_msg="error",
                            log_path="path",
                        ).msg,
                        code=messages.UNEXPECTED_ERROR.name,
                        meta={"service": "ent1"},
                    ),
                ],
            ),
        ),
    )
    @mock.patch(M_PATH + "_is_attached")
    @mock.patch(M_PATH + "_enabled_services")
    @mock.patch(M_PATH + "_reboot_required")
    @mock.patch(M_PATH + "attach_with_token")
    def test_failed_services_during_attach(
        self,
        m_attach_with_token,
        m_reboot_required,
        m_enabled_services,
        m_is_attached,
        _m_lock_enter,
        test_exception,
        expected_result,
        expected_warnings,
    ):
        m_reboot_required.return_value = mock.MagicMock(reboot_required="yes")
        m_is_attached.return_value = mock.MagicMock(is_attached=False)
        m_ent = mock.MagicMock()
        type(m_ent).name = mock.PropertyMock(return_value="ent3")
        m_enabled_services.return_value = mock.MagicMock(
            enabled_services=[m_ent]
        )
        m_attach_with_token.side_effect = test_exception

        actual_result = _full_token_attach(
            options=mock.MagicMock(token="token", auto_enable_services=True),
            cfg=None,
        )

        assert expected_result == actual_result
        assert expected_warnings == actual_result.warnings



from uaclient import (
    apt_news,
    config,
    entitlements,
    event_logger,
    exceptions,
    http,
    messages,
)
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.apt import AptProxyScope
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.files import state_files
from uaclient.files.user_config_file import LXDGuestAttachEnum
from uaclient.livepatch import (
    configure_livepatch_proxy,
    unconfigure_livepatch_proxy,
)
from uaclient.snap import configure_snap_proxy, unconfigure_snap_proxy

event = event_logger.get_event_logger()


def action_config(args, *, cfg, **kwargs):
    # Avoiding a circular import
    from uaclient.cli import get_parser

    get_parser().print_help_for_command("config")
    return 0


def action_config_show(args, *, cfg, **kwargs):
    """Perform the 'config show' action optionally limit output to a single key

    :return: 0 on success
    :raise UbuntuProError: on invalid keys
    """
    if args.key:  # limit reporting config to a single config key
        if args.key not in config.UA_CONFIGURABLE_KEYS:
            raise exceptions.InvalidArgChoice(
                arg="'{}'".format(args.key),
                choices=", ".join(config.UA_CONFIGURABLE_KEYS),
            )
        print(
            "{key} {value}".format(
                key=args.key, value=getattr(cfg, args.key, None)
            )
        )
        return 0

    col_width = str(max([len(x) for x in config.UA_CONFIGURABLE_KEYS]) + 1)
    row_tmpl = "{key: <" + col_width + "} {value}"

    for key in config.UA_CONFIGURABLE_KEYS:
        print(row_tmpl.format(key=key, value=getattr(cfg, key, None)))

    if (cfg.global_apt_http_proxy or cfg.global_apt_https_proxy) and (
        cfg.ua_apt_http_proxy or cfg.ua_apt_https_proxy
    ):
        print(messages.CLI_CONFIG_GLOBAL_XOR_UA_PROXY)


@cli_util.assert_root
def action_config_set(args, *, cfg, **kwargs):
    """Perform the 'config set' action.

    @return: 0 on success, 1 otherwise
    """
    from uaclient.cli import get_parser

    parser = get_parser()
    try:
        set_key, set_value = args.key_value_pair.split("=")
    except ValueError:
        parser.print_help_for_command("config set")
        raise exceptions.GenericInvalidFormat(
            expected="<key>=<value>", actual=args.key_value_pair
        )
    if set_key not in config.UA_CONFIGURABLE_KEYS:
        parser.print_help_for_command("config set")
        raise exceptions.InvalidArgChoice(
            arg="<key>", choices=", ".join(config.UA_CONFIGURABLE_KEYS)
        )
    if not set_value.strip():
        parser.print_help_for_command("config set")
        raise exceptions.EmptyConfigValue(arg=set_key)

    if type(getattr(cfg, set_key, None)) == bool:
        if set_value.lower() not in ("true", "false"):
            raise exceptions.InvalidArgChoice(
                arg="<value>", choices="true, false"
            )
        set_value = set_value.lower() == "true"

    if set_key in ("http_proxy", "https_proxy"):
        protocol_type = set_key.split("_")[0]
        if protocol_type == "http":
            validate_url = http.PROXY_VALIDATION_SNAP_HTTP_URL
        else:
            validate_url = http.PROXY_VALIDATION_SNAP_HTTPS_URL
        http.validate_proxy(protocol_type, set_value, validate_url)

        kwargs = {set_key: set_value}
        configure_snap_proxy(**kwargs)
        # Only set livepatch proxy if livepatch is enabled
        entitlement = entitlements.livepatch.LivepatchEntitlement(cfg)
        livepatch_status, _ = entitlement.application_status()
        if livepatch_status == ApplicationStatus.ENABLED:
            configure_livepatch_proxy(**kwargs)
    elif set_key in cfg.ua_scoped_proxy_options:
        protocol_type = set_key.split("_")[2]
        if protocol_type == "http":
            validate_url = http.PROXY_VALIDATION_APT_HTTP_URL
        else:
            validate_url = http.PROXY_VALIDATION_APT_HTTPS_URL
        http.validate_proxy(protocol_type, set_value, validate_url)
        unset_current = bool(
            cfg.global_apt_http_proxy or cfg.global_apt_https_proxy
        )
        if unset_current:
            print(
                messages.WARNING_APT_PROXY_OVERWRITE.format(
                    current_proxy="pro scoped apt", previous_proxy="global apt"
                )
            )
        cli_util.configure_apt_proxy(
            cfg, AptProxyScope.UACLIENT, set_key, set_value
        )
        cfg.global_apt_http_proxy = None
        cfg.global_apt_https_proxy = None

    elif set_key in (
        cfg.deprecated_global_scoped_proxy_options
        + cfg.global_scoped_proxy_options
    ):
        # setup_apt_proxy is destructive for unprovided values. Source complete
        # current config values from uaclient.conf before applying set_value.

        protocol_type = "https" if "https" in set_key else "http"
        if protocol_type == "http":
            validate_url = http.PROXY_VALIDATION_APT_HTTP_URL
        else:
            validate_url = http.PROXY_VALIDATION_APT_HTTPS_URL

        if set_key in cfg.deprecated_global_scoped_proxy_options:
            print(
                messages.WARNING_CONFIG_FIELD_RENAME.format(
                    old="apt_{}_proxy".format(protocol_type),
                    new="global_apt_{}_proxy".format(protocol_type),
                )
            )
            set_key = "global_" + set_key

        http.validate_proxy(protocol_type, set_value, validate_url)

        unset_current = bool(cfg.ua_apt_http_proxy or cfg.ua_apt_https_proxy)

        if unset_current:
            print(
                messages.WARNING_APT_PROXY_OVERWRITE.format(
                    current_proxy="global apt", previous_proxy="pro scoped apt"
                )
            )
        cli_util.configure_apt_proxy(
            cfg, AptProxyScope.GLOBAL, set_key, set_value
        )
        cfg.ua_apt_http_proxy = None
        cfg.ua_apt_https_proxy = None

    elif set_key in (
        "update_messaging_timer",
        "metering_timer",
    ):
        try:
            set_value = int(set_value)
            if set_value < 0:
                raise ValueError("Invalid interval for {}".format(set_key))
        except ValueError:
            parser.print_help_for_command("config set")
            # More readable in the CLI, without breaking the line in the logs
            print("")
            raise exceptions.InvalidPosIntConfigValue(
                key=set_key, value=set_value
            )
    elif set_key == "apt_news":
        if set_value:
            apt_news.update_apt_news(cfg)
        else:
            state_files.apt_news_contents_file.delete()
    elif set_key == "lxd_guest_attach":
        if not _is_attached(cfg).is_attached:
            raise exceptions.UnattachedError()
        state_files.lxd_pro_config_file.write(
            state_files.LXDProConfig(
                guest_attach=LXDGuestAttachEnum.from_value(set_value.lower())
            )
        )

    setattr(cfg, set_key, set_value)


@cli_util.assert_root
def action_config_unset(args, *, cfg, **kwargs):
    """Perform the 'config unset' action.

    @return: 0 on success, 1 otherwise
    """
    from uaclient.cli import get_parser

    if args.key not in config.UA_CONFIGURABLE_KEYS:
        parser = get_parser()
        parser.print_help_for_command("config unset")
        raise exceptions.InvalidArgChoice(
            arg="<key>", choices=", ".join(config.UA_CONFIGURABLE_KEYS)
        )
    if args.key in ("http_proxy", "https_proxy"):
        protocol_type = args.key.split("_")[0]
        unconfigure_snap_proxy(protocol_type=protocol_type)
        # Only unset livepatch proxy if livepatch is enabled
        entitlement = entitlements.livepatch.LivepatchEntitlement(cfg)
        livepatch_status, _ = entitlement.application_status()
        if livepatch_status == ApplicationStatus.ENABLED:
            unconfigure_livepatch_proxy(protocol_type=protocol_type)
    elif args.key in cfg.ua_scoped_proxy_options:
        cli_util.configure_apt_proxy(
            cfg, AptProxyScope.UACLIENT, args.key, None
        )
    elif args.key in (
        cfg.deprecated_global_scoped_proxy_options
        + cfg.global_scoped_proxy_options
    ):
        if args.key in cfg.deprecated_global_scoped_proxy_options:
            protocol_type = "https" if "https" in args.key else "http"
            event.info(
                messages.WARNING_CONFIG_FIELD_RENAME.format(
                    old="apt_{}_proxy".format(protocol_type),
                    new="global_apt_{}_proxy".format(protocol_type),
                )
            )
            args.key = "global_" + args.key
        cli_util.configure_apt_proxy(cfg, AptProxyScope.GLOBAL, args.key, None)
    elif args.key == "lxd_guest_attach":
        state_files.lxd_pro_config_file.write(
            state_files.LXDProConfig(guest_attach=LXDGuestAttachEnum.OFF)
        )

    setattr(cfg, args.key, None)
    return 0


show_subcommand = ProCommand(
    "show",
    help=messages.CLI_CONFIG_SHOW_DESC,
    description=messages.CLI_CONFIG_SHOW_DESC,
    action=action_config_show,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "key", help=messages.CLI_CONFIG_SHOW_KEY, nargs="?"
                )
            ]
        )
    ],
)

set_subcommand = ProCommand(
    "set",
    help=messages.CLI_CONFIG_SET_DESC,
    description=messages.CLI_CONFIG_SET_DESC,
    action=action_config_set,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "key_value_pair",
                    help=(
                        messages.CLI_CONFIG_SET_KEY_VALUE.format(
                            options=", ".join(config.UA_CONFIGURABLE_KEYS)
                        )
                    ),
                )
            ]
        )
    ],
)

unset_subcommand = ProCommand(
    "unset",
    help=messages.CLI_CONFIG_UNSET_DESC,
    description=messages.CLI_CONFIG_UNSET_DESC,
    action=action_config_unset,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "key",
                    help=(
                        messages.CLI_CONFIG_UNSET_KEY.format(
                            options=", ".join(config.UA_CONFIGURABLE_KEYS)
                        )
                    ),
                    metavar="key",
                )
            ]
        )
    ],
)

config_command = ProCommand(
    "config",
    help=messages.CLI_ROOT_CONFIG,
    description=messages.CLI_CONFIG_DESC,
    action=action_config,
    help_category=HelpCategory.OTHER,
    subcommands=[show_subcommand, set_subcommand, unset_subcommand],
)



import logging
from enum import Enum
from functools import lru_cache
from typing import Dict, Optional, Tuple, Type  # noqa: F401

from uaclient import clouds, exceptions, messages, system, util
from uaclient.config import apply_config_settings_override

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

CLOUD_TYPE_TO_TITLE = {
    "aws": "AWS",
    "aws-china": "AWS China",
    "aws-gov": "AWS Gov",
    "azure": "Azure",
    "gce": "GCP",
}

PRO_CLOUD_URLS = {
    "aws": messages.urls.PRO_ON_AWS_HOME_PAGE,
    "azure": messages.urls.PRO_ON_AZURE_HOME_PAGE,
    "gce": messages.urls.PRO_ON_GCP_HOME_PAGE,
}


class NoCloudTypeReason(Enum):
    NO_CLOUD_DETECTED = 0
    CLOUD_ID_ERROR = 1


def get_instance_id() -> Optional[str]:
    """Query cloud instance-id from cmdline."""
    try:
        # Present in cloud-init on >= Xenial
        out, _err = system.subp(["cloud-init", "query", "instance_id"])
        return out.strip()
    except exceptions.ProcessExecutionError:
        pass
    LOG.warning("Unable to determine current instance-id")
    return None


@lru_cache(maxsize=None)
@apply_config_settings_override("cloud_type")
def get_cloud_type() -> Tuple[Optional[str], Optional[NoCloudTypeReason]]:
    if system.which("cloud-id"):
        # Present in cloud-init on >= Xenial
        try:
            out, _err = system.subp(["cloud-id"])
            return (out.strip(), None)
        except exceptions.ProcessExecutionError as exc:
            LOG.debug("error running cloud-id: %s", str(exc))
            return (None, NoCloudTypeReason.CLOUD_ID_ERROR)
    # If no cloud-id command, assume not on cloud
    return (None, NoCloudTypeReason.NO_CLOUD_DETECTED)


def cloud_instance_factory(
    cloud_override: Optional[str] = None,
) -> clouds.AutoAttachInstance:
    """
    :raises CloudFactoryError: if no cloud instance object can be constructed
    :raises CloudFactoryNoCloudError: if no cloud instance object can be
        constructed because we are not on a cloud
    :raises CloudFactoryUnsupportedCloudError: if no cloud instance object can
        be constructed because we don't have a class for the cloud we're on
    :raises CloudFactoryNonViableCloudError: if no cloud instance object can be
        constructed because we explicitly do not support the cloud we're on
    """
    from uaclient.clouds import aws, azure, gcp, lxd

    cloud_instance_map = {
        "aws": aws.AWSAutoAttachInstance,
        "aws-china": aws.AWSAutoAttachInstance,
        "aws-gov": aws.AWSAutoAttachInstance,
        "azure": azure.AzureAutoAttachInstance,
        "gce": gcp.GCPAutoAttachInstance,
        "lxd": lxd.LXDAutoAttachInstance,
    }  # type: Dict[str, Type[clouds.AutoAttachInstance]]

    if cloud_override is not None:
        cloud_type = cloud_override
    else:
        cloud_type, _ = get_cloud_type()

    if not cloud_type:
        raise exceptions.CloudFactoryNoCloudError()
    cls = cloud_instance_map.get(cloud_type)
    if not cls:
        raise exceptions.NonAutoAttachImageError(cloud_type=cloud_type)
    instance = cls()
    if not instance.is_viable:
        raise exceptions.CloudFactoryNonViableCloudError()
    return instance



from uaclient import exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.apt import is_installed
from uaclient.clouds.identity import cloud_instance_factory
from uaclient.config import UAConfig
from uaclient.data_types import BoolDataValue, DataObject, Field


class ShouldAutoAttachResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "should_auto_attach",
            BoolDataValue,
            doc="True if the system should run auto-attach on boot",
        ),
    ]

    def __init__(self, should_auto_attach: bool):
        self.should_auto_attach = should_auto_attach


def should_auto_attach() -> ShouldAutoAttachResult:
    return _should_auto_attach(UAConfig())


def _should_auto_attach(cfg: UAConfig) -> ShouldAutoAttachResult:
    """
    This endpoint checks if a given system should run auto-attach on boot.
    """
    try:
        cloud_instance_factory()
    except exceptions.CloudFactoryError:
        return ShouldAutoAttachResult(
            should_auto_attach=False,
        )

    return ShouldAutoAttachResult(
        should_auto_attach=is_installed("ubuntu-advantage-pro")
        or is_installed("ubuntu-pro-auto-attach"),
    )


endpoint = APIEndpoint(
    version="v1",
    name="ShouldAutoAttach",
    fn=_should_auto_attach,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.11",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.attach.auto.should_auto_attach.v1 import should_auto_attach

result = should_auto_attach()
""",  # noqa: E501
    "result_class": ShouldAutoAttachResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.attach.auto.should_auto_attach.v1",
    "example_json": """
{
    "should_auto_attach": false
}
""",
}



import mock
import pytest

from uaclient.api.u.pro.version.v1 import VersionError, VersionResult, _version
from uaclient.testing import fakes


class TestVersionV1:
    @mock.patch("uaclient.api.u.pro.version.v1.get_version", return_value="28")
    def test_version(self, _m_get_version, FakeConfig):
        result = _version(cfg=FakeConfig())
        assert isinstance(result, VersionResult)
        assert result.installed_version == "28"

    @mock.patch("uaclient.api.u.pro.version.v1.get_version")
    def test_version_error(self, m_get_version, FakeConfig):
        m_get_version.side_effect = fakes.FakeUbuntuProError()

        with pytest.raises(VersionError) as excinfo:
            _version(FakeConfig())

        assert excinfo.errisinstance(VersionError)
        assert (
            excinfo.value.msg == "Unable to determine version: This is a test"
        )
        assert excinfo.value.msg_code == "unable-to-determine-version"



from behave import when

from features.steps.contract import (
    change_contract_endpoint_to_production,
    change_contract_endpoint_to_staging,
)
from features.steps.shell import when_i_run_command


@when("I fix `{issue}` by attaching to a subscription with `{token_type}`")
def when_i_fix_a_issue_by_attaching(context, issue, token_type):
    token = getattr(context.pro_config, token_type)

    if (
        token_type == "contract_token_staging"
        or token_type == "contract_token_staging_expired"
    ):
        change_contract_endpoint_to_staging(context, user_spec="with sudo")
    else:
        change_contract_endpoint_to_production(context, user_spec="with sudo")

    when_i_run_command(
        context=context,
        command="pro fix {}".format(issue),
        user_spec="with sudo",
        stdin="a\n{}\n".format(token),
        verify_return=False,
    )


@when("I fix `{issue}` by enabling required service")
def when_i_fix_a_issue_by_enabling_service(context, issue):
    when_i_run_command(
        context=context,
        command="pro fix {}".format(issue),
        user_spec="with sudo",
        stdin="e\n",
    )


@when("I fix `{issue}` by updating expired token")
def when_i_fix_a_issue_by_updating_expired_token(context, issue):
    token = getattr(context.pro_config, "contract_token")
    when_i_run_command(
        context=context,
        command="pro fix {}".format(issue),
        user_spec="with sudo",
        stdin="r\n{}\n".format(token),
    )



import time

from uaclient import exceptions, secret_manager
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.contract import UAContractClient
from uaclient.data_types import (
    DataObject,
    Field,
    IntDataValue,
    StringDataValue,
)

MAXIMUM_ATTEMPTS = 70


class MagicAttachWaitOptions(DataObject):
    fields = [
        Field(
            "magic_token",
            StringDataValue,
            doc="The Token provided by the initiate endpoint.",
        ),
    ]

    def __init__(self, magic_token: str):
        self.magic_token = magic_token


class MagicAttachWaitResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "user_code",
            StringDataValue,
            doc=(
                "Code the user will see in the UI when confirming the Magic"
                " Attach"
            ),
        ),
        Field(
            "token",
            StringDataValue,
            doc="The same Magic Token that was sent as an argument",
        ),
        Field(
            "expires",
            StringDataValue,
            doc="Timestamp of the Magic Attach process expiration",
        ),
        Field(
            "expires_in",
            IntDataValue,
            doc="Seconds before the Magic Attach process expires",
        ),
        Field(
            "contract_id",
            StringDataValue,
            doc="ID of the contract the machine will be attached to",
        ),
        Field(
            "contract_token",
            StringDataValue,
            doc="The contract Token to attach the machine",
        ),
    ]

    def __init__(
        self,
        user_code: str,
        token: str,
        expires: str,
        expires_in: int,
        contract_id: str,
        contract_token: str,
    ):
        self.user_code = user_code
        self.token = token
        self.expires = expires
        self.expires_in = expires_in
        self.contract_id = contract_id
        self.contract_token = contract_token


def wait(
    options: MagicAttachWaitOptions,
) -> MagicAttachWaitResult:
    return _wait(options, UAConfig())


def _wait(
    options: MagicAttachWaitOptions, cfg: UAConfig
) -> MagicAttachWaitResult:
    """
    This endpoint polls the contracts service waiting for the user to confirm
    the Magic Attach.
    """
    contract = UAContractClient(cfg)

    num_attempts = 0
    num_connection_errors = 0
    wait_time = 10

    secret_manager.secrets.add_secret(options.magic_token)
    while num_attempts < MAXIMUM_ATTEMPTS:
        wait_resp = None

        try:
            wait_resp = contract.get_magic_attach_token_info(
                magic_token=options.magic_token
            )
            num_connection_errors = 0
        except exceptions.MagicAttachTokenError:
            break
        # If the server is unavailable we will bump the wait
        # time. We will return to the normal amount if we can
        # successfully reach the server and we verify that
        # the contractToken information is still not being
        # returned
        except exceptions.MagicAttachUnavailable:
            wait_time = 30
        # If we have a flaky connectivity error, this part of the code
        # will make sure that we try at least three more times before
        # raising a ConnectivityError.
        except exceptions.ConnectivityError as e:
            if num_connection_errors < 3:
                num_connection_errors += 1
            else:
                raise e

        if wait_resp and wait_resp.get("contractToken") is not None:
            return MagicAttachWaitResult(
                user_code=wait_resp["userCode"],
                token=wait_resp["token"],
                expires=wait_resp["expires"],
                expires_in=int(wait_resp["expiresIn"]),
                contract_id=wait_resp["contractID"],
                contract_token=wait_resp["contractToken"],
            )
        elif wait_resp:
            wait_time = 10

        time.sleep(wait_time)
        num_attempts += 1

    raise exceptions.MagicAttachTokenError()


endpoint = APIEndpoint(
    version="v1",
    name="MagicAttachWait",
    fn=_wait,
    options_cls=MagicAttachWaitOptions,
)

_doc = {
    "introduced_in": "27.11",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.attach.magic.wait.v1 import MagicAttachWaitOptions, wait

options = MagicAttachWaitOptions(magic_token="<magic_token>")
result = wait(options)
""",  # noqa: E501
    "result_class": MagicAttachWaitResult,
    "exceptions": [
        (
            exceptions.ConnectivityError,
            (
                "Raised if it is not possible to connect to the contracts"
                " service."
            ),
        ),
        (
            exceptions.ContractAPIError,
            (
                "Raised if there is an unexpected error in the contracts"
                " service interaction."
            ),
        ),
        (
            exceptions.MagicAttachTokenError,
            "Raised when an invalid/expired Token is sent.",
        ),
        (
            exceptions.MagicAttachUnavailable,
            (
                "Raised if the Magic Attach service is busy or unavailable at"
                " the moment."
            ),
        ),
    ],
    "example_cli": "pro api u.pro.attach.magic.wait.v1 --args magic_token=<magic_token>",  # noqa: E501
    "example_json": """
{
    "user_code":"<UI_code>",
    "token":"<magic_token>",
    "expires": "<yyyy-MM-dd>T<HH:mm:ss>.<TZ>",
    "expires_in": 500,
    "contract_id": "<Contract-ID>",
    "contract_token": "<attach_token>",
}
""",
}



"""
Generate RST documentation for an ubuntu-pro-client api endpoint.

For a single endpoint: python3 tools/apidocgen.py $ENDPOINT_NAME
For complete API reference documentation: python3 tools/apidocgen.py
"""

import sys
import textwrap
from importlib import import_module

sys.path.insert(0, ".")

from uaclient import data_types  # noqa: E402
from uaclient.api.api import VALID_ENDPOINTS  # noqa: E402
from uaclient.exceptions import UbuntuProError  # noqa: E402

BLOCK_INDENT = "   "
LIST_INDENT = "  "
NO_ARGS = "- This endpoint takes no arguments."
NO_FIELDS = "- This object has no fields."

ENDPOINT_TEMPLATE = """\
{name}
{underline}

{description}

- Introduced in Ubuntu Pro Client Version: ``{introduced_in}~``
- Requires network access: {requires_network}

Arguments:

{args_section}

.. tab-set::

   .. tab-item:: Python API interaction
      :sync: python

      **Calling from Python code:**

      .. code-block:: python

         {example_python}

      **Expected return object:**

      {result_classes}

      **Raised exceptions:**

      {exceptions}

   .. tab-item:: CLI interaction
      :sync: CLI

      **Calling from the CLI:**

      .. code-block:: bash

         {example_cli}

      {example_cli_extra}

      **Expected attributes in JSON structure:**

      .. code-block:: js

         {example_json}

{extra}

"""


def data_value_type_to_str(data_value_cls):
    if data_value_cls.__name__ == "_DataList":
        return "List[{}]".format(
            data_value_type_to_str(data_value_cls.item_cls)
        )
    if issubclass(data_value_cls, data_types.DataObject):
        return data_value_cls.__name__
    return data_value_cls.python_type_name


def create_fields_table(cls):
    table_template = """
.. list-table::
   :header-rows: 1

   * - Field Name
     - Type
     - Description
{fields}"""
    field_definition_template = """
* - ``{name}``
  - ``{type}``
  - {description}"""

    if len(cls.fields) == 0:
        return NO_FIELDS

    fields = []
    for field in cls.fields:
        type_str = data_value_type_to_str(field.data_cls)
        if not field.required:
            type_str = "Optional[{}]".format(type_str)
        fields.append(
            textwrap.indent(
                field_definition_template.format(
                    name=field.key, type=type_str, description=field.doc or ""
                ).strip(),
                prefix=BLOCK_INDENT,
            )
        )
    return table_template.format(
        fields="\n".join(fields),
    ).strip()


def collect_data_object_classes(target_cls, collection, *, ignore=[]):
    for field in target_cls.fields:
        cls = field.data_cls
        if cls.__name__ == "_DataList":
            cls = cls.item_cls
        if cls.__name__ == "_DataDict":
            cls = cls.dict_value_cls
        if (
            issubclass(cls, data_types.DataObject)
            and cls not in collection
            and cls not in ignore
        ):
            collection.append(cls)
            collect_data_object_classes(cls, collection, ignore=[])


def print_endpoint_docs(endpoint_name):
    module = import_module("uaclient.api." + endpoint_name)

    if module.endpoint.options_cls is None:
        args_section = NO_ARGS
    else:
        args_section = create_fields_table(module.endpoint.options_cls)
    extra_args_content = module._doc.get("extra_args_content")
    if extra_args_content:
        args_section += "\n\n" + extra_args_content.strip()

    result_class = module._doc.get("result_class")
    result_classes = [result_class]
    collect_data_object_classes(
        result_class,
        result_classes,
        ignore=module._doc.get("ignore_result_classes", []),
    )
    result_classes += module._doc.get("extra_result_classes", [])
    class_definition_template = """
- ``{module}.{name}``

{fields_table}
    """
    result_class_definitions = []
    for cls in result_classes:
        result_class_definitions.append(
            class_definition_template.format(
                module=cls.__module__,
                name=cls.__qualname__,
                fields_table=textwrap.indent(
                    create_fields_table(cls), LIST_INDENT
                ),
            ).strip()
        )
    result_class_definitions_str = textwrap.indent(
        "\n\n".join(result_class_definitions),
        prefix=BLOCK_INDENT * 2,
    ).strip()

    possible_exceptions = [
        (
            UbuntuProError,
            (
                "``UbuntuProError`` is the base class of all exceptions raised"
                " by Ubuntu Pro Client and it is best practice to handle this"
                " error on any API call. (Note: if any API call raises an"
                " exception that does not inherit from ``UbuntuProError``,"
                " please report a bug)."
            ),
        )
    ] + module._doc.get("exceptions", [])
    exception_str_list = []
    for err_cls, msg in possible_exceptions:
        exception_str_list.append(
            "- ``{name}``: {msg}".format(name=err_cls.__name__, msg=msg)
        )
    exceptions_str = textwrap.indent(
        "\n".join(exception_str_list), prefix=BLOCK_INDENT * 2
    ).strip()

    extra_indent = module._doc.get("extra_indent", 1)

    data = {
        "name": endpoint_name,
        "underline": "=" * len(endpoint_name),
        "description": textwrap.dedent(module.endpoint.fn.__doc__).strip(),
        "introduced_in": module._doc.get("introduced_in"),
        "requires_network": "Yes" if module._doc["requires_network"] else "No",
        "args_section": args_section,
        "example_python": textwrap.indent(
            module._doc.get("example_python").strip(),
            prefix=BLOCK_INDENT * 3,
        ).strip(),
        "result_classes": result_class_definitions_str,
        "exceptions": exceptions_str,
        "example_cli": module._doc.get("example_cli").strip(),
        "example_cli_extra": textwrap.indent(
            module._doc.get("example_cli_extra", ""),
            prefix=BLOCK_INDENT * 2,
        ).strip(),
        "example_json": textwrap.indent(
            module._doc.get("example_json").strip(),
            prefix=BLOCK_INDENT * 3,
        ).strip(),
        "extra": textwrap.indent(
            module._doc.get("extra", "").strip(),
            prefix=BLOCK_INDENT * extra_indent,
        ),
    }

    print(ENDPOINT_TEMPLATE.format(**data))


if __name__ == "__main__":
    if len(sys.argv) > 1:
        # assume arg is endpoint name
        endpoint_name = sys.argv[1]
        try:
            print_endpoint_docs(endpoint_name)
        except ModuleNotFoundError:
            print(f"Endpoint `{endpoint_name}` does not exist")
            sys.exit(1)
    else:
        # all
        print(
            """\
..
   THIS DOCUMENTATION WAS AUTOMATICALLY GENERATED
   Do not edit this document directly. Instead, edit the metadata in the api
   source code. The api source code can be found on the main branch of the git
   repo at https://github.com/canonical/ubuntu-pro-client. The api source code
   is nested in the uaclient/api folder. If you need to edit the documentation
   for u.pro.version.v1 you can find the api source code in
   uaclient/api/u/pro/version/v1.py The documentation metadata is at the bottom
   of that file.
"""
        )
        print()
        print("Available endpoints")
        print("===================")
        print()
        print("The currently available endpoints are:")
        print()
        for e in VALID_ENDPOINTS:
            print(f"- `{e}`_")
        print()
        for e in VALID_ENDPOINTS:
            try:
                print_endpoint_docs(e)
            except Exception:
                msg = f"Failed generating docs for {e}"
                print(msg)
                print("^" * len(msg))
                raise



import mock
import pytest

from uaclient import exceptions
from uaclient.api.u.pro.attach.auto.should_auto_attach.v1 import (
    _should_auto_attach,
)

M_PATH = "uaclient.api.u.pro.attach.auto.should_auto_attach.v1"


class TestShouldAutoAttachV1:
    @pytest.mark.parametrize(
        "expected,installed_pkgs",
        (
            (True, ("foo", "ubuntu-advantage-pro", "bar")),
            (False, ("foo", "bar")),
        ),
    )
    @mock.patch(M_PATH + ".cloud_instance_factory")
    @mock.patch("uaclient.apt.get_installed_packages_names")
    def test_detect_is_pro(
        self,
        m_get_installed_pkgs,
        _m_cloud_factory,
        expected,
        installed_pkgs,
        FakeConfig,
    ):
        m_get_installed_pkgs.return_value = installed_pkgs
        assert expected == _should_auto_attach(FakeConfig()).should_auto_attach

    @pytest.mark.parametrize(
        "cloud_exception",
        (
            (exceptions.CloudFactoryNoCloudError()),
            (exceptions.NonAutoAttachImageError(cloud_type="test")),
            (exceptions.CloudFactoryNonViableCloudError()),
        ),
    )
    @mock.patch(M_PATH + ".cloud_instance_factory")
    def test_detect_is_pro_when_cloud_factory_fails(
        self,
        m_cloud_factory,
        cloud_exception,
        FakeConfig,
    ):
        m_cloud_factory.side_effect = cloud_exception
        assert False is _should_auto_attach(FakeConfig()).should_auto_attach
